# -*- coding: utf-8 -*-
"""Component for predicting the gender from a BAM alignment file.

Accepts the path to a BAM file and optionally a prepend for the id of
the sex chromosomes.

The component reads coverage for subsections of each sex chromosome.
Based on the ratio between the average coverage across chromosomes it
makes a simple yet very accurate gender prediction.
"""
from __future__ import absolute_import

from .cli import sex_check
from .core import Gender, gender_from_bam, predict_gender



# -*- coding: utf-8 -*-
"""
chanjo.builder.core
~~~~~~~~~~~~~~~~~~~~

Central pipeline for the Chanjo builder module.
"""
from __future__ import absolute_import
import errno

from path import path
from toolz import pipe, reduce, concat
from toolz.curried import map

from .._compat import text_type
from .consumers import commit_per_contig
from .stages import aggregate, build_block, build_interval, build_superblock
from ..utils import bed_to_interval, split


def init_db(chanjo_db, bed_stream, overwrite=False):
  """Build a new database instance from the Chanjo BED stream.

  Args:
    chanjo_db (Store): initialized Store class instance
    bed_stream (sequence): Chanjo-style BED-stream
    overwrite (bool, optional): whether to automatically overwrite an
      existing database, defaults to False
  """
  # check if the database already exists (expect 'mysql' to exist)
  # 'dialect' is in the form of '<db_type>+<connector>'
  if chanjo_db.dialect == 'mysql' or path(chanjo_db.uri).exists():
    if overwrite:
      # wipe the database clean with a warning
      chanjo_db.tare_down()
    elif chanjo_db.dialect == 'sqlite':
      # prevent from wiping existing database to easily
      raise OSError(errno.EEXIST, chanjo_db.uri)

  # set up new tables
  chanjo_db.set_up()

  superblocks = pipe(
    bed_stream,
    map(text_type.rstrip),
    map(split(sep='\t')),
    map(lambda row: bed_to_interval(*row)),
    map(build_interval(chanjo_db)),
    concat,
    aggregate,
    map(build_block(chanjo_db)),
    aggregate,
    map(build_superblock(chanjo_db))
  )

  # reduce the superblocks and commit every contig
  reduce(commit_per_contig(chanjo_db), superblocks, 'chr0')

  # commit also the last contig
  chanjo_db.save()



# -*- coding: utf-8 -*-
from codecs import open

from chanjo.converter import ccds_to_bed


def test_builder_pipeline():
  """Test the entire converter pipeline."""
  # test with minimal BED "file"
  ccds_stream = open('tests/fixtures/CCDS.mini.txt', 'r', encoding='utf-8')
  intervals = list(ccds_to_bed(ccds_stream))

  assert len(intervals) == 19

  # test interval belonging to two blocks
  special_interval = intervals[3]
  assert special_interval.block_ids == ['CCDS54521.1', 'CCDS46694.1']
  assert special_interval.superblock_ids == ['RFPL2', 'RFPL2']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import click

from .core import init_pipeline
from .. import __banner__


@click.command()
@click.option('-r', '--remove', is_flag=True, help='remove config option')
@click.argument('key')
@click.argument('value', required=False)
@click.pass_context
def config(context, key, value, remove):
  """Only handles string values."""
  # update or delete the config key-value pair
  # the key can be provide as a path to nested commands: "user.name"
  if remove:
    context.obj.unset(key, scope=context.obj.user_data)

  else:
    if value is None:
      raise ValueError("Unless remove a setting, you must submit a 'value'")

    if value.isnumeric():
      value = int(value)

    context.obj.set(key, value, scope=context.obj.user_data)

  # persist updates to the config file
  context.obj.save()


@click.command()
@click.pass_context
def init(context):
  """Walk user through setting up a new config file."""
  # print a nice welcome message
  click.echo(__banner__)

  questions = [
    ('annotate.cutoff', 'sufficient coverage',
      context.obj.get('annotate', {}).get('cutoff', 10)),
    ('dialect', 'preferred SQL-dialect', context.parent.db.dialect),
    ('db', 'central database path/URI', context.parent.db.uri)
  ]

  # launch init pipeline
  init_pipeline(__package__, context.obj, questions)



# -*- coding: utf-8 -*-
from toolz import curry


@curry
def build_interval_data(chanjo_db, sample_id, group_id, interval_data):
  # create a new intervals data entry
  return chanjo_db.add(chanjo_db.create(
    'interval_data',
    parent_id=interval_data[3],
    sample_id=sample_id,
    group_id=group_id,
    coverage=float(interval_data[-2]),      # second to last field
    completeness=float(interval_data[-1])   # last field
  ))



# -*- coding: utf-8 -*-
"""
chanjo.utils
~~~~~~~~~~~~~

A few general utility functions that might also be useful outside
Chanjo.
"""
from __future__ import absolute_import, division, unicode_literals
from collections import namedtuple
import random

from toolz import concat, curry

from ._compat import text_type


_RawInterval = namedtuple('RawInterval', [
  'contig', 'start', 'end', 'name', 'score', 'strand',
  'block_ids', 'superblock_ids', 'coverage', 'completeness'
])


class BaseInterval(_RawInterval):

  """Immutable interval tuple based on the `BED format`_.

  Args:
    contig (str): chromosome or generic contig id
    start (int): chromosomal start position, 1-based
    end (int): chromosomal end position, 1-based
    name (str, optional): unique name/id of interval
    score (int, optional): value between 0-1000, not used in Chanjo
    strand (str, optional): +/-
    block_ids (list, optional): list of unique block ids
    superblock_ids (list, optional): list of unique superblock ids
    coverage (float, optional): average coverage across the interval
    completeness (float, optional): completeness across the interval

  .. _BED format: http://genome.ucsc.edu/FAQ/FAQformat.html#format1
  """

  def __new__(cls, contig, start, end, name='', score='', strand='',
              block_ids=None, superblock_ids=None,
              coverage='', completeness=''):
    """Initialize a new namedtuple instance."""
    return super(BaseInterval, cls).__new__(
      cls,
      contig,
      start,
      end,
      name,
      score,
      strand,
      (block_ids or []),       # default to empty lists
      (superblock_ids or []),  # do
      coverage,
      completeness
    )


def bed_to_interval(contig, bed_start, bed_end, name='', score='', strand='',
                    block_ids='', superblock_ids=''):
  """Convert from a BED row to an immutable BaseInterval object.

  Args:
    contig (str): chromosome or generic contig id
    start (str): chromosomal start position, 1-based
    end (str): chromosomal end position, 1-based
    name (str, optional): unique name/id of interval
    score (str, optional): value between 0-1000, not used in Chanjo
    strand (str, optional): +/-
    block_ids (str, optional): list of unique related block ids
    superblock_ids (str, optional): list of unique superblock ids

  Returns:
    namedtuple: processed and sanity checked interval object
  """
  try:
    # assure positions to be integers
    # convert from 0,1-based to 1,1-based positions
    start = int(bed_start) + 1
    end = int(bed_end)
  except ValueError:
    raise ValueError("'start' and 'end' should be integers")

  # perform sanity check to check for incorrect formatting
  assert (end - start) >= 0, ("Not a valid BED interval."
                              "(bedEnd - bedStart) must be >= 0.")

  # fallback to empty list for optional element ids
  ids = [element_ids.split(',') if element_ids else []
         for element_ids in (block_ids, superblock_ids)]

  return BaseInterval(contig, start, end, name, score, strand, *ids)


def average(sequence):
  """Calculate the mean across an array of e.g. read depths.

  Defaults to the mean calculated using numpy and falls back to the
  naive Python solution.

  Args:
    sequence (list): ``numpy.array`` or list of values

  Returns:
    float: calculated average value
  """
  try:
    # first assume that numpy is installed for the fastest approach
    return sequence.mean()

  except AttributeError:
    # no numpy available, fall back to support regular list
    return sum(sequence) / len(sequence)


def completeness(read_depths, threshold=10):
  """Calculate completeness across a range of read depths.

  Note:
    This function catches the corner case where ``read_depths`` is an
    empty array which would lead to a ``ZeroDivisionError`` and returns
    0% by default.

  Args:
    read_depths (array): :class:`numpy.array` of read depths for
      **each** of the positions
    threshold (int, optional): cutoff to use for the filter

  Returns:
    float: calculated completeness in percent

  Examples:

  .. code-block:: python

    >>> completeness([5, 6, 6, 7, 6, 5, 5], threshold=6)
    0.571428571
  """
  base_pair_count = len(read_depths)

  # dodge rare division by zero error when `read_depths` is an empty array
  try:
    # filter, then count bases with greater read depth than `threshold`
    # divide by the total number of bases
    return len(read_depths[read_depths >= threshold]) / base_pair_count

  except ZeroDivisionError:
    # without any bases to check, 0% of bases passed the threshold
    return 0.


@curry
def serialize_interval(interval, delimiter='\t', subdelimiter=',', bed=False):
  r"""Stringify :class:`BaseInterval`.

  Args:
    interval (:class:`BaseInterval`): interval object to serialize
    delimiter (str, optional): main delimiter, defaults to "\t"
    subdelimiter (str, optional): secondary delimiter, defaults to ","
    bed (bool, optional): convert to BED interval (0:1-based)

  Returns:
    str: stringified version of the :class:`BaseInterval`

  Examples:

  .. code-block:: python

    >>> interval = BaseInterval('chr1', 10, 100, score=14)
    >>> serialize_interval(interval)
    'chr1\t10\t100\t\t14'
  """
  # serialize the list of block/superblock Ids
  block_ids = subdelimiter.join(interval.block_ids)
  superblock_ids = subdelimiter.join(interval.superblock_ids)

  # on request, convert positions from 1:1 to BED-style 0:1
  if bed:
    base = interval._replace(start=interval.start - 1)[:6]
  else:
    base = interval[:6]

  return text_type.rstrip(
    delimiter.join(
      # concat and stringify base, related ids, and (possible) metrics
      map(text_type, concat([base, [block_ids, superblock_ids], interval[8:]]))
    ),
    delimiter   # strip trailing delimiters
  )


def id_generator(size=8):
  """Randomly generate an id of length N (size) we can recognize.

  Think Italian or Japanese or Native American.
  Modified from: `Stackoverflow <http://stackoverflow.com/questions/2257441>`_
  and `ActiveState <http://code.activestate.com/recipes/526619/>`_.

  Args:
    size (int, optional): length of id, number of characters

  Returns:
    str: randomly generated string

  Examples:

  .. code-block:: python

    >>> id_generator()
    'palevedu'
    >>> id_generator(3)
    'sun'
  """
  variables = 'aeiou'
  consonants = 'bdfghklmnprstvw'

  return ''.join(
    [random.choice(variables if i % 2 else consonants) for i in range(size)]
  )


@curry
def split(string, sep='\t'):
  r"""Split string based on a delimiter/separator.

  This wrapper around ``str.split`` harmonize Python 2/3 syntax.

  Args:
    string (str): string to "explode"
    sep (str, optional): delimiter to separate on. Default: \t.

  Returns:
    list: "exploded" string parts as a list

  Examples:

  .. code-block:: python

    >>> split('change|of|pace', sep='|')
    ['change', 'of', 'pace']
  """
  return text_type.split(string, sep)



# -*- coding: utf-8 -*-
"""
chanjo.builder
~~~~~~~~~~~~~~~

Module for initializing a new instance of a Chanjo SQL database.

Accepts a Chanjo-BED stream as input defining intervals and how they
relate to blocks and superblocks. You can think about these as the
eqivalent of exons, transcripts, and genes. It's required that the
input BED stream is sorted by start position and chromosome for the
function to work.

The component build a base SQL database of genomic elements. It also
sets up foreign keys between elements. There's nothing to pipe to a
post-process.

Usage:

.. code-block:: python

  >>> from chanjo import builder
  >>> with open('intervals.sorted.bed', 'r') as stream:
  ...   # use the BED stream to build a new SQLite database
  ...   # overwrite any potentially existing database
  ...   builder.init_db(stream, './coverage.sqlite', overwrite=True)

"""
from __future__ import absolute_import

from .cli import build
from .core import init_db
from .consumers import commit_per_contig
from .stages import aggregate, build_block, build_interval, build_superblock



# -*- coding: utf-8 -*-
from __future__ import division, unicode_literals

from chanjo.annotator import annotate_bed_stream
from chanjo.utils import BaseInterval


def test_annotator_pipeline():
  """Test the entire annotator pipeline."""
  bam_path = 'tests/fixtures/alignment.bam'

  # test with minimal BED "file"
  bed_stream = ['#chrom\tstart\tend', '1\t0\t5', '1\t9\t20']
  read_depths1 = [2, 4, 5, 5, 5]
  read_depths2 = [7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7]
  result = annotate_bed_stream(
    bed_stream, bam_path, cutoff=5, contig_prefix='chr', bp_threshold=100
  )

  interval1 = next(result)
  interval2 = next(result)

  assert interval1[:3] == BaseInterval('chr1', 1, 5)[:3]
  assert interval1.coverage == sum(read_depths1) / len(read_depths1)
  assert interval1.completeness == 3 / 5

  assert interval2[:3] == BaseInterval('chr1', 10, 20)[:3]
  assert interval2.coverage == sum(read_depths2) / len(read_depths2)
  assert interval2.completeness == 1.



# -*- coding: utf-8 -*-
from codecs import open

from chanjo.builder import init_db
from chanjo.store import Store


def test_builder_pipeline():
  """Test the entire builder pipeline."""
  chanjo_db = Store(':memory:')
  chanjo_db.set_up()

  # test with minimal BED "file"
  bed_stream = open('tests/fixtures/CCDS.mini.bed', 'r', encoding='utf-8')
  init_db(chanjo_db, bed_stream)

  block = chanjo_db.get('block', 'CCDS2.2')
  assert block.start == 12
  assert block.end == 35

  superblock = chanjo_db.get('superblock', 'RFPL2')
  assert superblock.start == 32586759
  assert superblock.end == 32589260

  block_ids = [block_.id for block_ in superblock.blocks]
  assert set(block_ids) == set(['CCDS46694.1', 'CCDS54521.1'])

  assert len(chanjo_db.find('block')) == 5
  assert len(chanjo_db.find('superblock')) == 4



# -*- coding: utf-8 -*-
"""
chanjo.converter
~~~~~~~~~~~~~~~~~

Module for converting a CCDS dump to standard Chanjo BED-format.
"""
from __future__ import absolute_import

from .cli import convert
from .core import ccds_to_bed
from .stages import (
  grep,
  extract_intervals,
  merge_related_elements,
  parse_raw_intervals,
  rename_sex_interval
)



# -*- coding: utf-8 -*-


def convert_old_interval_id(old_id):
  """Deprecated function for converting a 0:0-based exon Id to the
  new 1:1-based interval Id.

  Args:
    old_id (str): Old exon Id, '0:0-based'

  Returns:
    str: New interval Id, '1:1-based'
  """
  # Split into parts (contig, start, end)
  parts = old_id.split('-')

  # Recombine but with converted coordinates from 0:0 to 1:1
  return '-'.join([parts[0], str(int(parts[1]) + 1), str(int(parts[2]) + 1)])



# -*- coding: utf-8 -*-
"""
chanjo.converter.stages
~~~~~~~~~~~~~~~~~~~~~~~~

Pipeline stages only used by the Chanjo converter.
"""
from __future__ import absolute_import
from operator import attrgetter

from toolz import curry, mapcat

from ..utils import BaseInterval


@curry
def grep(pattern, string):
  """Match a simple pattern substring in a given string.

  Note that the function would also work to check for an item in a list,
  a key in a dictionary etc.

  Args:
    pattern (str): substring to match with
    string (str): string to match against

  Returns:
    bool: if ``pattern`` was a substring in ``string``
  """
  return pattern in string


def parse_raw_intervals(str_list):
  """Decode serialized CCDS exons.

  Accepts a formatted string of interval coordinates from the CCDS
  row and turns it into a more manageable list of lists with
  (start, end) coordinates for each interval (exon).

  .. code-block:: python

    >>> parse_raw_intervals('[11-18, 25-30, 32-35]')
    [[11, 18], [25, 30], [32, 35]]

  Args:
    str_list (str): A CSV string of (start, end) pairs, wrapped in '[]'

  Returns:
    list: 2D list with the start ``int``, end ``int`` pairs
  """
  # remove the "[]"
  csv_intervals = str_list[1:-1].replace(' ', '')

  # 1. split first into exons coordinates
  # 2. split into start, end and parse int
  intervals = [[int(pos) for pos in item.split('-')]
               for item in csv_intervals.split(',')]

  return intervals


def extract_intervals(record):
  """Compile an BaseInterval from a single (split) CCDS record row.

  Args:
    record (tuple): split CCDS row

  Yields:
    BaseInterval: namedtuple class representation of an interval
  """
  # extract contig Id as string
  contig = record[0]

  # parse the intervals list-string and yield each of the intervals
  for start, end in parse_raw_intervals(record[9]):

    yield BaseInterval(
      contig,                             # contig
      start,                              # start
      end,                                # end
      "%s-%d-%d" % (contig, start, end),  # unique id
      0,                                  # score, unused but required
      record[6],                          # strand
      [record[4]],                        # block ids
      [record[2]]                         # superblock ids
    )


def rename_sex_interval(interval, sex_contigs=('X', 'Y')):
  """Rename interval ids for intervals on sex chromosomes.

  Doesn't do anything but return non-sex interval.

  The need for this step is that 20 superblocks are present on both sex
  chromosomes (X and Y). However, corresponding intervals should still
  be treated as if they wheren't really homologos.

  Args:
    interval (tuple): tuple representation of an interval

  Returns:
    BaseInterval: namedtuple representation of an interval
  """
  contig = interval.contig
  if contig in sex_contigs:
    # keep the funtion pure, avoid altering input object!
    return interval._replace(
      block_ids=["%s-%s" % (contig, block_id)
                 for block_id in interval.block_ids],
      superblock_ids=["%s-%s" % (contig, superblock_id)
                      for superblock_id in interval.superblock_ids]
    )

  else:
    return interval


def merge_related_elements(interval_group):
  """Merge block and superblock ids for a group of identical intervals.

  Args:
    interval_group (list): list of identical intervals

  Returns:
    BaseInterval: unified interval with merged related element ids
  """
  # extract block and superblock ids from each of the intervals
  block_ids = mapcat(attrgetter('block_ids'), interval_group)
  superblock_ids = mapcat(attrgetter('superblock_ids'), interval_group)

  return BaseInterval(
    *interval_group[0][:6],              # use first interval as base
    block_ids=list(block_ids),           # resolve and add
    superblock_ids=list(superblock_ids)  # do
  )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from pkg_resources import resource_filename, resource_listdir

import click
from path import path

@click.command()
@click.argument(
  'location', type=click.Path(), default='./chanjo-demo', required=False)
@click.pass_context
def demo(context, location):
  """Copy demo files to a directory.

  \b
  LOCATION: directory to add demofiles to (default: ./chanjo-demo)
  """
  user_dir = path(location)
  demo_dir = path(resource_filename(__package__, 'demo'))

  # make sure we don't overwrite exiting files
  for demo_file in resource_listdir(__package__, 'demo'):
    user_file_path = user_dir.joinpath(demo_file)
    if user_file_path.exists():
      click.echo(user_file_path + ' exists. Pick a different location.')
      context.abort()

  try:
    # we can copy the directory(tree)
    demo_dir.copytree(user_dir)
  except OSError:
    click.echo('The location must be a non-existing directory.')
    context.abort()

  # inform the user
  click.echo("Successfully copied demo files to %s." % user_dir)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import click

from .._compat import text_type
from .core import gender_from_bam


@click.command('sex-check')
@click.option(
  '-p', '--prefix', default='', help='prefix a string to each contig')
@click.argument('bam_path', type=click.Path(exists=True))
@click.version_option('1.0')
def sex_check(bam_path, prefix):
  """Sex Check - predict gender from a BAM-alignment.

  \b
  BAM_PATH: path to BAM-file
  """
  # run the sex checker pipeline
  gender = gender_from_bam(bam_path, prefix=prefix)

  # print the results to the console for pipeability (csv)
  click.echo("#%(prefix)sX_coverage\t%(prefix)sY_coverage\tsex"
             % dict(prefix=prefix))
  click.echo('\t'.join(map(text_type, gender)))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division
from collections import namedtuple

from toolz import pipe
from toolz.curried import map

from ..depth_reader import BamFile
from ..utils import average

Gender = namedtuple('Gender', ['x_coverage', 'y_coverage', 'sex'])


def predict_gender(x_coverage, y_coverage):
  """Make a simple yet accurate prediction of a samples gender.

  The calculation is based on the average coverage across the X and Y
  chromosomes. Note that an extrapolation from a subsection of bases is
  usually quite sufficient.

  Args:
    x_coverage (float): estimated average coverage on X chromosome
    y_coverage (float): estimated average coverage on Y chromosome

  Returns:
    str: prediction, ['male', 'female', 'unknown']
  """
  # algoritm doesn't work if coverage is missing for X chromosome
  if x_coverage == 0:
    return 'unknown'

  # this is the entire prediction, it's usually very obvious
  elif (y_coverage > 0) and (x_coverage / y_coverage < 10):
    return 'male'

  else:
    # the few reads mapping to the Y chromosomes are artifacts
    return 'female'


def gender_from_bam(bam_path, prefix=''):
  """Predict the gender from a BAM alignment file.

  Args:
    bam_path (path): path to a BAM alignment file
    prefix (str, optional): string to prefix to 'X', 'Y'

  Returns:
    Gender: tuple of X coverage, Y coverage, and sex prediction

  Examples:
    >>> gender_from_bam('alignment.bam', prefix='chr')
    Gender(x_coverage=123.31, y_coverage=0.13, sex='female')
  """
  # setup: connect to a BAM file
  bam = BamFile(bam_path)

  # step 0: fake some BED interval rows (already 1,1-based!)
  fake_bed_rows = [("%sX" % prefix, 1, 59373566),
                   ("%sY" % prefix, 69362, 11375310)]

  # step 1: run the pipeline
  sequence = pipe(
    fake_bed_rows,
    map(lambda interval: bam(*interval)),
    map(average)
  )

  # step: make the prediction
  x_coverage, y_coverage = list(sequence)
  sex = predict_gender(x_coverage, y_coverage)
  return Gender(x_coverage, y_coverage, sex)



# -*- coding: utf-8 -*-
"""
chanjo.annotator.core
~~~~~~~~~~~~~~~~~~~~~~

Central pipeline for the Chanjo annotator module.
"""
from __future__ import absolute_import, unicode_literals

from toolz import concat, pipe
from toolz.curried import complement, filter, map

from .._compat import text_type
from ..depth_reader import BamFile
from .stages import (
  calculate_metrics,
  comment_sniffer,
  extend_interval,
  group_intervals,
  prefix,
  process_interval_group
)
from ..utils import bed_to_interval, split


def annotate_bed_stream(bed_stream, bam_path, cutoff=10, extension=0,
                        contig_prefix='', bp_threshold=17000):
  """Annotate all intervals from a BED-file stream.

  Yields tuple data for each interval with calculated coverage and
  completeness.

  Args:
    bed_stream (sequence): usually a BED-file handle to read from
    bam_path (str): path to BAM-file
    cutoff (int, optional): threshold for completeness calculation,
      defaults to 10
    extension (int, optional): number of bases to extend each interval
      with (+/-), defaults to 0
    contig_prefix (str, optional): rename contigs by prefixing,
      defaults to empty string
    bp_threshold (int, optional): optimization threshold for reading
      BAM-file in chunks, default to 17000

  Yields:
    tuple: :class:`chanjo.BaseInterval`, coverage (float), and
      completeness (float)
  """
  # setup: connect to BAM-file
  bam = BamFile(bam_path)

  # the pipeline
  return pipe(
    bed_stream,
    filter(complement(comment_sniffer)),         # filter out comments
    map(text_type.rstrip),                       # strip invisble chars.
    map(prefix(contig_prefix)),                  # prefix to contig
    map(split(sep='\t')),                        # split lines
    map(lambda row: bed_to_interval(*row)),      # convert to objects
    map(extend_interval(extension=extension)),   # extend intervals
    group_intervals(bp_threshold=bp_threshold),  # group by threshold
    map(process_interval_group(bam)),            # read coverage
    concat,                                      # flatten list of lists
    map(calculate_metrics(threshold=cutoff))     # calculate cov./compl.
  )



# -*- coding: utf-8 -*-
from chanjo.store import Interval, Interval_Block


def test_Interval_Block():
  # test table name
  assert Interval_Block.name == 'interval__block'


def test_Interval():
  # set up
  interval = Interval('interval1', 'chr1', 10, 1000, '-')

  assert len(interval) == 991



#!/usr/bin/env python
from __future__ import division

import numpy as np
import pytest

from chanjo._compat import text_type
from chanjo.utils import (
  average,
  _RawInterval,
  bed_to_interval,
  completeness,
  id_generator,
  BaseInterval,
  serialize_interval
)

def test_RawInterval():
  """Test generating a base :class:`_RawInterval`."""
  interval = (
    'chr1', 10, 20, 'int1', 0, '-', ['block1'], ['superblock1'], 12.3, .93
  )
  bed_interval = _RawInterval(*interval)

  assert bed_interval == interval
  assert bed_interval.start == 10
  assert bed_interval.contig == 'chr1'
  assert bed_interval.block_ids == ['block1']
  assert bed_interval.completeness == .93


def test_BaseInterval():
  """Test generating an interval without all fields filled in."""
  interval = ('chr1', 10, 20, 'int1')
  bed_interval = BaseInterval(*interval)

  assert bed_interval != interval
  assert bed_interval.start == 10
  assert bed_interval.contig == 'chr1'
  assert bed_interval.score == ''
  assert bed_interval.name == 'int1'
  assert bed_interval.block_ids == []


def test_bed_to_interval():
  """Test converting between BED interval and Chanjo interval."""
  bed_interval = ('22', '101', '220', 'int32', 0, '+', 'block11')
  chanjo_interval = bed_to_interval(*bed_interval)

  assert chanjo_interval.contig == '22'
  # BED specifies 0:1 intervals, Chanjo works with 1:1 mapping
  assert chanjo_interval.start == 102
  assert chanjo_interval.end == 220
  assert chanjo_interval.block_ids == ['block11']
  assert chanjo_interval.superblock_ids == []

  with pytest.raises(ValueError):
    # interval with invalid coordinates/arguments
    bed_to_interval(20, 'X', 24, 'int2', 0, '-')


def test_average():
  """Test calculating average of a list of values."""
  values = [0, 5, 5, 6]
  # 'without' numpy
  assert average(values) == 4.

  # with numpy array
  assert average(np.array(values)) == 4.


def test_completeness():
  """Test calculating completeness of a list of read depths."""
  # test simple case
  assert completeness(np.array([0, 10, 10, 20, 10, 0])) == 4/6

  # test edge case with 0 positions, should *not* raise
  # ``ZeroDivisionError``.
  assert completeness(np.array([])) == 0.

  # test with a different ``threshold``
  assert completeness(np.array([20, 40, 10, 0, 10]), threshold=30) == 1/5


def test_serialize_interval():
  """Test serializing an BaseInterval instance."""
  # simple case, should remove empty fields to the right
  interval = BaseInterval('chr1', 10, 20)
  assert serialize_interval(interval) == 'chr1\t10\t20'

  # with block ids, should maintain empty intermediate fields!
  interval = BaseInterval('chr22', 101, 200, block_ids=['block11', 'block12'])
  serialized_interval = 'chr22\t101\t200\t\t\t\tblock11,block12'
  assert serialize_interval(interval) == serialized_interval

  # test curried function composition
  serialize_interval_alt = serialize_interval(delimiter='|', subdelimiter=';')
  serialized_interval_alt = 'chr22|101|200||||block11;block12'
  assert serialize_interval_alt(interval) == serialized_interval_alt


def test_id_generator():
  """Test generating a random id."""
  # difficult to test randomly generated strings...
  assert len(id_generator()) == 8
  assert isinstance(id_generator(), text_type)

  # test with custom sized id
  assert len(id_generator(3)) == 3

  # test edge case with 0 lenght
  assert id_generator(0) == ''



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from codecs import open
import json
import toml

from click.utils import LazyFile
import pytest

from chanjo.config import Config


def test_init():
  # Test lazy loading (not pointing to a config file)
  config = Config()
  assert config.markup is json

  # Test loading with some defaults
  config = Config(defaults=dict(name='P.T. Anderson'))
  assert config['name'] == 'P.T. Anderson'

  # Test loading with an alternate markup engine
  config = Config(markup=toml)
  assert config.markup is toml

  # Test loading with a non-existent config file
  config = Config(LazyFile('tests/fixtures/nothing.json', 'w'))
  assert config == {}

  # Test loading with an existing config file
  config = Config(LazyFile('tests/fixtures/something.json', 'w'))
  assert config['title'] == 'Magnolia'


class TestConfig:
  def setup(self):
    self.defaults = {
      'type': 'movie',
      'person': {'name': 'P.T. Anderson', 'job': 'Director'}
    }
    self.config = Config(defaults=self.defaults)

  def test_load(self):
    # Test loading non-conflicting data
    with open('tests/fixtures/something.json', encoding='utf-8') as handle:
      self.config.load(handle)

    assert self.config['person']['name'] == 'P.T. Anderson'
    assert self.config['title'] == 'Magnolia'

    # Test loading conflicting data
    with open('tests/fixtures/conflict.json', encoding='utf-8') as handle:
      self.config.load(handle)

    assert self.config['name'] == 'Alfonso Cuarón'

    # Test loading with a non-supported markup engine
    self.config.markup = dict()
    with pytest.raises(NotImplementedError):
      with open('tests/fixtures/something.json', encoding='utf-8') as handle:
        self.config.load(handle)

    # Test loading config with bad syntax
    self.config.markup = json
    with pytest.raises(ValueError):
      with open('tests/fixtures/bad-syntax.json', encoding='utf-8') as handle:
        self.config.load(handle)

  def test_resolve_key(self):
    # Test simple case
    section, key = self.config._resolve_key('person.name')
    assert section == self.defaults['person']
    assert key == 'name'

    # Test with a different base
    section, key = self.config._resolve_key('person.job',
                                            base=self.config.user_data)
    assert section == {}  # 'user_data' is an empty ``dict`` in this case
    assert key == 'job'

  def test_set(self):
    # Test setting a global setting
    self.config.set('country', 'United States')
    assert self.config['country'] == 'United States'

    # Test setting a nested setting (in existing sub-group)
    self.config.set('person.name', 'Quentin Tarantino')
    assert self.config['person']['name'] == 'Quentin Tarantino'

    # Test setting a nested setting (in a new sub-group)
    self.config.set('bonus.joke', 'Knock, knock.')
    assert self.config['bonus']['joke'] == 'Knock, knock.'

    # Test multiple nested settings
    value = 'Do not ask me'
    self.config.set('what.is.this.useful.for', value)
    assert self.config['what']['is']['this']['useful']['for'] == value

    # Test set on a different scope (than ``self``)
    self.config.set('nested.key', 'value', scope=self.config.user_data)
    assert self.config.user_data['nested']['key'] == 'value'

  def test_unset(self):
    # Test removing a global setting
    assert self.config['type'] == 'movie'
    self.config.unset('type')
    assert self.config.get('type') is None

    # Test removing a nested setting
    assert self.config['person']['name'] == 'P.T. Anderson'
    self.config.unset('person.name')
    assert self.config['person'].get('name') is None

    # Test removing a whole sub-group of settings
    assert self.config['person'] is not None
    self.config.unset('person')
    assert self.config.get('person') is None

    # Test removing a setting in a different scope
    scope = {'person': {'name': 'Tarantino'}}
    self.config.unset('person.name', scope=scope)
    assert scope['person'].get('name') is None



# -*- coding: utf-8 -*-
"""
chanjo
~~~~~~~

Coverage analysis for clinical sequencing. It's intuatively documented
and embraces functional programming patterns.

:copyright: (c) 2014 by Robin Andeer
:licence: MIT, see LICENCE for more details
"""
from __future__ import absolute_import, unicode_literals
from .annotator import annotate_bed_stream
from .builder import init_db
from .converter import ccds_to_bed
from .exporter import export_intervals
from .importer import import_bed, import_json
from .sex_checker import gender_from_bam
from .store import Store
from .utils import (
  average,
  BaseInterval,
  bed_to_interval,
  completeness,
  id_generator,
  serialize_interval,
  split
)

__banner__ = r"""
       ______               ________
 _________  /_______ _____________(_)_____
 _  ___/_  __ \  __ `/_  __ \____  /_  __ \
 / /__ _  / / / /_/ /_  / / /___  / / /_/ /   by Robin Andeer
 \___/ /_/ /_/\__,_/ /_/ /_/___  /  \____/
                            /___/
"""

__title__ = 'chanjo'
__summary__ = 'coverage analysis tool for clinical sequencing'
__uri__ = 'http://www.chanjo.co/'

__version__ = '2.0.0-beta'
__codename__ = 'Wistful Weasel'

__author__ = 'Robin Andeer'
__email__ = 'robin.andeer@gmail.com'

__license__ = 'MIT'
__copyright__ = 'Copyright 2014 Robin Andeer'



# -*- coding: utf-8 -*-
"""
chanjo.builder.consumers
~~~~~~~~~~~~~~~~~~~~~~~~~

*Reduce* friendly functions to consume a sequence of items.
"""
from toolz import curry


@curry
def commit_per_contig(chanjo_db, prev_contig, next_superblock):
  """Commit freshly created elements to the database.

  Built up to be used as the reducing function in the Python ``reduce``
  function.

  .. code-block:: python

    >>> reduce(commit_per_contig(chanjo_db), superblocks, 'chr0')

  Args:
    chanjo_db (Store): Chanjo store connected to an existing database
    prev_contig (str): contig id of the previous superblock
    next_superblock (str): contig id of the next/current superblock
  """
  # commit once for every finished contig
  if prev_contig != next_superblock.contig:
    chanjo_db.save()

  return next_superblock.contig



# -*- coding: utf-8 -*-
"""
chanjo.questions
~~~~~~~~~~~~~~~~~~
"""
from __future__ import absolute_import, unicode_literals

from click.termui import echo, style

# Python 2.x compatability
try:
  input = raw_input
except NameError:
  pass


def build_prompt(prompt, replacement=''):
  """Craft a prompt to ask a question.

  Provides a few modifications that should make it as simple as possible
  to ask questions.

  .. code-block:: python

    >>> build_prompt('name', '(Heisenberg)')
    'name: (Heisenberg)'
    >>> build_prompt('What is your name?', '(Heisenberg)')
    'What is your name? (Heisenberg)'
    >>> build_prompt('I %sed to China', '[style(walk, fg='red')]')
    'I [walk]ed to China'

  Args:
    prompt (str): Base prompt, ``%s`` will be substituted with 'replacement'
    replacement (str, optional): String to substitute ``%s`` with

  Returns:
    str: The full, modified prompt string
  """
  # prefix the question-prefix
  prompt = ("[%s] " % style('?', fg='green')) + prompt

  # determine whether it seems user has tried to format the prompt
  if '%s' not in prompt:
    # automate some formatting for the sake of convenience
    if not prompt.endswith(' '):
      if not prompt.endswith('?') or prompt.endswith(':'):
        prompt += ':'

    # add the default option to the end of prompt unless specified elsewhere
    prompt += " %s"

  # make the default-substitution and ensure an empty space at the end
  return (prompt % replacement).rstrip() + ' '


def ask(prompt, default=None, color='cyan'):
  """Ask a question, waits for user input.

  Replacement for "input". Updates the same line by replacing "default".

  .. code-block:: python

    >>> my_name = ask('Say my name: %s ', 'Heisenberg')
    # Wait for user input
    Say my name: (Heisenberg) Walter
    # Updates *the same* line with 'Walter' in green
    Say my name: Walter
    >>> print(my_name)
    Walter

  Inspired by 'bower_ init' which confirms user input by replacing the
  default option in line (ref_).

  Args:
    prompt (str): Question to print to user, '%s' will be replaced by default
    default (str, optional): Default option unless replaced by user
    color (str, optional): Some common color like 'red', 'green', 'yellow'

  Returns:
    str: User input or default

  .. _bower: http://bower.io/
  .. _ref: http://stackoverflow.com/questions/12586601
  """
  # helper variables
  MOVE_CURSOR_UP = '\x1b[1A'
  ERASE_LINE = '\x1b[2K'

  # determine if a default was submitted
  if default:
    # prepare the default-part of the prompt
    default_string = "(%s)" % default
  else:
    # not relevant since ``promt`` shouldn't include a '%s'
    default_string = ''

  # pass question to user and wait for response
  # write default option in parentheses, use it as response if nothing
  # was submitted by user.
  response = input(build_prompt(prompt, default_string)) or default

  # print the updated confirmation line by replacing the previous
  echo(MOVE_CURSOR_UP + ERASE_LINE
       + build_prompt(prompt, style(str(response) or '', fg=color)))

  return response


def questionnaire(questions, confirm_color='cyan'):
  """Ask a set of questions á la 'bower_ init'.

  Returns a dict with responses (or defaults). Abstraction for multiple
  consecutive calls to :func:`ask`.

  Note that the function doesn't try to be smart about what types
  (e.g. 'int' or 'str') the user is typing in. You should do any
  nessesary conversions downstream.

  .. code-block:: python

    >>> questions = [('name', 'Heisenberg'), ('age', 'How old are you?', 50)]
    >>> questionnaire(questions)
    name: (Heisenberg) Walter       # -> name: Walter
    How old are you? (25)           # -> How old are you? 25
    [out] {'age': 25, 'name': 'Walter'}

  Args:
    questions (list): List of tuples on the form:
      ``(<question id>, [<prompt>], [<default>])``.
    confirm_color str, optional): Some common color like to color the
      confirm messages.

  Returns:
    dict: Each question Id is returned with either user input or the
      default value (**unmodified**)

  .. _bower: http://bower.io/
  """
  # initialize dict to store responses (or defaults)
  responses = {}

  # ask each question to sending to :func:`ask`
  for question in questions:
    # save responses with question Ids
    responses[question[0]] = ask(*question[-2:], color=confirm_color)

  return responses



# -*- coding: utf-8 -*-
from chanjo.sex_checker import Gender, gender_from_bam, predict_gender


def test_Gender():
  # create a gender tuple
  gender = Gender(10.2, 8.1, 'male')

  assert gender.x_coverage == 10.2
  assert gender.y_coverage == 8.1
  assert gender.sex == 'male'


def test_predict_gender():
  assert predict_gender(x_coverage=10.2, y_coverage=8.1) == 'male'
  assert predict_gender(x_coverage=20.9, y_coverage=.01) == 'female'
  # shouldn't raise ``ZeroDivisionError``
  assert predict_gender(x_coverage=5.12, y_coverage=0) == 'female'

  # what does this even mean?
  assert predict_gender(x_coverage=0, y_coverage=10) == 'unknown'



# -*- coding: utf-8 -*-
from chanjo.builder import (
  aggregate, build_block, build_interval, build_superblock
)
from chanjo.utils import BaseInterval
from chanjo.store import Store, Interval as StoreInterval


def test_aggregate():
  payloads = [
    ('block1', BaseInterval('X', 10, 100)),
    ('block1', BaseInterval('X', 50, 130)),
    ('block2', BaseInterval('Y', 111, 222))
  ]

  groups = [
    [('block1', BaseInterval('X', 10, 100)),
     ('block1', BaseInterval('X', 50, 130))],
    [('block2', BaseInterval('Y', 111, 222))]
  ]

  assert groups == list(aggregate(payloads))


class TestBuildElement(object):
  def setup(self):
    self.store = Store(':memory:')
    self.store.set_up()

    self.interval = BaseInterval(
      '1', 10, 100,
      name='int1',
      block_ids=['block1', 'block2'],
      superblock_ids=['sblock1', 'sblock1']
    )

    self.db_interval = StoreInterval(
      interval_id=self.interval.name,
      contig=self.interval.contig,
      start=self.interval.start,
      end=self.interval.end,
      strand=self.interval.strand
    )
    self.interval_group = [('block2', self.db_interval, 'sblock1')]

  def test_build_interval(self):
    intervals = list(build_interval(self.store, self.interval))
    self.store.save()
    new_interval = self.store.get('interval', self.interval.name)

    assert new_interval.id == self.interval.name
    assert new_interval.start == self.interval.start

    assert len(intervals) == 2

  def test_build_block(self):
    superblock_id, new_block = build_block(self.store, self.interval_group)
    self.store.save()
    db_block = self.store.get('block', self.interval_group[0][0])

    assert superblock_id == 'sblock1'
    assert new_block.id == db_block.id == self.interval_group[0][0]
    assert new_block.start == db_block.start == self.interval_group[0][1].start



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import click

from .core import import_bed, import_json


@click.command(name='import')
@click.option(
  '-j', '--json', is_flag=True, help="use legacy JSON 'annotate' output")
@click.argument(
  'in_stream', type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def import_data(context, in_stream, json):
  """Import coverage annotations to an existing database.

  \b
  IN_STREAM: Chanjo-style BED-file with interval definitions
  """
  args = (context.parent.db, in_stream)

  if json:
    # FYI: the ``bed_stream`` really is a JSON-file
    import_json(*args)
  else:
    import_bed(*args)



# -*- coding: utf-8 -*-
"""
chanjo.annotator
~~~~~~~~~~~~~~~~~

This module is responsible for the annotation of coverage and
completeness of intervals in BED-file.

Exposes the central functions and pipeline found deeper in the module.
"""
from __future__ import absolute_import

from .core import annotate_bed_stream
from .stages import (
  assign_relative_positions,
  calculate_metrics,
  comment_sniffer,
  extend_interval,
  group_intervals,
  merge_intervals,
  prefix,
  process_interval_group
)



# -*- coding: utf-8 -*-
"""
chanjo.converter.core
~~~~~~~~~~~~~~~~~~~~~~

Central pipeline for the Chanjo converter module.
"""
from __future__ import absolute_import
from itertools import groupby as lazy_groupby
from operator import attrgetter

from toolz import pipe, partial, concat
from toolz.curried import map, filter, groupby, valmap, pluck

from .._compat import itervalues, text_type
from .stages import (
  extract_intervals, grep, merge_related_elements, rename_sex_interval
)
from ..utils import split


def ccds_to_bed(ccds_stream):
  """Convert CCDS dump to Chanjo-style BED stream.

  Main entry point for default Chanjo converter (ccds). It converts
  a sorted (start, chrom) CCDS database to the Chanjo BED-format.

  Args:
    ccds_stream (file): file handle to read CCDS lines from

  Yields:
    Interval: interval with merged block and superblock ids
  """
  return pipe(
    ccds_stream,
    filter(grep('Public')),                    # filter out Public tx
    map(text_type.rstrip),                     # strip \n and spaces
    map(split(sep='\t')),                      # split into list
    map(extract_intervals),                    # convert to Interval
    concat,                                    # flatten
    map(rename_sex_interval),                  # rename sex contigs
    partial(lazy_groupby, key=attrgetter('contig')),  # group by contig
    pluck(1),                                  # extract second item
    map(groupby(attrgetter('name'))),          # non-lazy group by id
    map(valmap(merge_related_elements)),       # group intervals
    map(itervalues),                           # extract values
    map(partial(sorted, key=attrgetter('start'))),  # sort by start pos
    concat                                     # flatten
  )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json

import click
from path import path
from toolz import pipe
from toolz.curried import map

from .core import annotate_bed_stream
from ..utils import id_generator, serialize_interval


@click.command()
@click.option('-s', '--sample', help='unique id (otherwise auto-generated)')
@click.option('-g', '--group', help='id to associate samples e.g. in trios')
@click.option('-c', '--cutoff', default=10, help='cutoff for completeness')
@click.option(
  '-e', '--extendby', default=0, help='extend intervals symetrically')
@click.option(
  '-p', '--prefix', default='', help='prefix a string to each contig')
@click.option(
  '-t', '--threshold',
  default=17000,
  help='base pair threshold to optimize BAM-file reading')
@click.argument('bam_path', type=click.Path(exists=True))
@click.argument(
  'in_stream', type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def annotate(context, bam_path, in_stream, sample, group, cutoff,
             extendby, prefix, threshold):
  """Annotate intervals in a BED-file/stream.

  \b
  BAM_PATH: Path to BAM-file
  IN_STREAM: Chanjo-style BED-file with interval definitions
  """
  # user defined sample id or randomly generated
  sample = (sample or id_generator())

  # step 1: metadata header
  metadata = dict(
    sample_id=sample,
    group_id=group,
    cutoff=cutoff,
    coverage_source=path(bam_path).abspath(),
    extension=extendby
  )
  click.echo("#%s" % json.dumps(metadata))

  # step 2: annotate list of intervals with coverage and completeness
  bed_lines = pipe(
    annotate_bed_stream(
      bed_stream=in_stream,
      bam_path=bam_path,
      cutoff=cutoff,
      extension=extendby,
      contig_prefix=prefix,
      bp_threshold=threshold
    ),
    map(serialize_interval)    # stringify/bedify
  )

  # reduce/write the BED lines
  for bed_line in bed_lines:
    click.echo(bed_line)



# -*- coding: utf-8 -*-
"""Component for exporting intervals in the Chanjo-BED format from an
existing Chanjo database.

The component only needs the path to the Chanjo database and performs
the nessesary SQL queries and formatting to transform each interval
to the Chanjo-flavoured BED-format.
"""
from __future__ import absolute_import

from .cli import export
from .core import export_intervals
from .producers import fetch_records



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from toolz import pipe, concat, juxt, last
from toolz.curried import compose, map, take

from .producers import fetch_records
from ..store import Interval


def export_intervals(chanjo_db, include_header=True, bed_score=0):
  r"""Return BED-formatted interval lines from existing ``chanjo_db``.

  BED lines are ready to be printed or written to a file.

  Args:
    chanjo_db (session): ``sqlalchemy.orm.session`` object with a
      ``.query``-method
    include_header (bool, optional): whether to include BED header
    bed_score (int, optional): dummy score (0-1000) to insert at field 5
      to complete the BED format

  Yields:
    str: stringified and tab-delimited interval

  Examples:
    >>> from chanjo import export_intervals, Store
    ... # instantiate a new connection to a Chanjo database
    >>> db = Store('./coverage.sqlite')
    >>> with open('intervals.sorted.bed', 'w') as stream:
    ...   # write intervals in BED-format with appropriate headers
    ...   for bed_line in export_intervals(db):
    ...     stream.write(bed_line + '\n')
  """
  if include_header:
    yield '#chrom\tchromStart\tchromEnd\tname\tscore\tstrand'

  # setup up which columns to fetch to make BED file
  # column 5 is just a silly default for the "score" field in BED
  i = Interval  # alias
  columns = (i.contig, i.start - 1, i.end, i.id, i.strand)

  # BED files are tab-delimited
  delimiter = '\t'

  # 1. fetch interval tuples from the database (producer)
  # 2. stringify each item in each subsequence (interval tuple)
  # 3. join lines on tab-character
  # 4. prepend the header
  bed_lines = pipe(
    fetch_records(chanjo_db, columns),
    map(map(str)),                        # convert fields to strings
    map(juxt(compose(list, take(4)),      # keep first 4 fields
             lambda _: [str(bed_score)],  # insert BED score
             compose(list, last))),       # keep last field
    map(concat),                          # flatten each item
    map(delimiter.join)                   # join on \t
  )

  for bed_line in bed_lines:
    yield bed_line



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from datetime import datetime

from sqlalchemy import (
  Table, Column, ForeignKey, String, Integer,
  DateTime, Text, Float, Boolean
)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declarative_base

# base for declaring a mapping
Base = declarative_base()

# +--------------------------------------------------------------------+
# | Association tables
# | ~~~~~~~~~~~~~~~~~~~
# | Provides the many-to-many relationships between:
# | - Interval<->Block
# +--------------------------------------------------------------------+
Interval_Block = Table('interval__block', Base.metadata,
  Column('interval_id', String(32), ForeignKey('interval.id')),
  Column('block_id', String(32), ForeignKey('block.id'))
)


# +--------------------------------------------------------------------+
# | Superblock ('gene') ORM
# +--------------------------------------------------------------------+
class Superblock(Base):

  """Collection of blocks and potentially overlapping intervals.

  A :class:`Superblock` can be related to multiple blocks and multiple
  intervals.

  Args:
    superblock_id (str): unique superblock id e.g. HGNC gene symbol
    contig (str): contig/chromosome id
    start (int): 1-based start of the superblock (first interval)
    end (int): 1-based end of the superblock (last interval, no UTR)
    strand (str): strand +/-
    secondary_id (str): e.g. Entrez gene id
  """

  __tablename__ = 'superblock'

  id = Column(String(32), primary_key=True)
  secondary_id = Column(String(32))
  contig = Column(String(5))
  start = Column(Integer)
  end = Column(Integer)
  strand = Column(String(1))

  def __init__(self, superblock_id=None, contig=None, start=None,
               end=None, strand=None, secondary_id=None):
    super(Superblock, self).__init__()

    self.id = superblock_id
    self.secondary_id = secondary_id
    self.contig = contig
    self.start = start
    self.end = end
    self.strand = strand


# +--------------------------------------------------------------------+
# | Block ('transcript') ORM
# +--------------------------------------------------------------------+
class Block(Base):

  """Set of non-overlapping intervals.

  A :class:`Block` can *only* be related to a single superset.

  Args:
    block_id (str): unique block id (e.g. CCDS transcript id)
    contig (str): contig/chromosome id
    start (int): 1-based start of the block (first interval)
    end (int): 1-based end of the block (last interval, no UTR)
    strand (str): strand +/-
    superblock_id (str): related superblock id, e.g. HGNC gene symbol
    secondary_id (str, optional): secondard block id
  """

  __tablename__ = 'block'

  id = Column(String(32), primary_key=True)
  secondary_id = Column(String(32))
  contig = Column(String(5))
  start = Column(Integer)
  end = Column(Integer)
  strand = Column(String(1))

  superblock_id = Column(String(32), ForeignKey('superblock.id'))
  superblock = relationship(
    Superblock, backref=backref('blocks', order_by=start))

  def __init__(self, block_id=None, contig=None, start=None, end=None,
               strand=None, superblock_id=None, secondary_id=None):
    super(Block, self).__init__()

    self.id = block_id
    self.contig = contig
    self.start = start
    self.end = end
    self.strand = strand
    self.superblock_id = superblock_id
    self.secondary_id = secondary_id

  def __len__(self):
    """Return the combined number of exon bases.

    Excludes intronic bases.

    Returns:
      int: total 'intervalic' (exonic) length of the block
    """
    base_count = 0
    for interval in self.intervals:
      base_count += len(interval)

    return base_count


# +--------------------------------------------------------------------+
# | Interval ('exon') ORM
# | TODO: rename this class to not be the same as named tuple!!
# +--------------------------------------------------------------------+
class Interval(Base):

  """A continous genetic interval on a given contig (e.g. exon).

  A :class:`Interval` can be related to a multiple :class:`Block`
  (transcripts). Start and end coordinates are 1-based.

  Args:
    interval_id (str): unique interval id
    contig (str): contig/chromosome id
    start (int): 1-based start of the interval
    end (int): 1-based end of the interval
    strand (str): strand +/-
  """

  __tablename__ = 'interval'

  id = Column(String(32), primary_key=True)
  contig = Column(String(5))
  start = Column(Integer)
  end = Column(Integer)
  strand = Column(String(1))
  first = Column(Boolean)
  last = Column(Boolean)

  # defines the ``backref`` to give transcripts an exons property
  blocks = relationship(Block, secondary=Interval_Block,
                        backref=backref('intervals', order_by=start))

  def __init__(self, interval_id, contig, start, end, strand):
    super(Interval, self).__init__()

    self.id = interval_id
    self.contig = contig
    self.start = start
    self.end = end
    self.strand = strand

  def __len__(self):
    """Return the number of bases.

    Returns:
      int: length of interval in number of bases
    """
    # add +1 because both coordinates are 1-based (number of *bases*)
    return (self.end - self.start) + 1


# +--------------------------------------------------------------------+
# | Sample ORM classes
# +--------------------------------------------------------------------+
class Sample(Base):

  """Metadata for a single (unique) sample.

  :class:`Sample` helps out in consolidating important information in
  one place.

  .. versionadded:: 0.4.0

  Args:
    sample_id (str): unique sample id
    group_id (str): unique group id
    cutoff (int): cutoff used for completeness
    extension (bool): number of bases added to each interval
    coverage_source (str): path to the BAM file used
  """

  __tablename__ = 'sample'

  id = Column(String(32), primary_key=True)
  group_id = Column(String(32), index=True)

  cutoff = Column(Integer)
  extension = Column(Integer)
  coverage_source = Column(Text)
  element_source = Column(Text)
  created_at = Column(DateTime, default=datetime.now)
  updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now)

  def __init__(self, sample_id, **kwargs):
    super(Sample, self).__init__()

    self.id = sample_id

    for key, value in kwargs.items():
      setattr(self, key, value)


# +--------------------------------------------------------------------+
# | Interval Data ORM
# +--------------------------------------------------------------------+
class IntervalData(Base):

  """Coverage metrics for a single interval and a given sample.

  :class:`IntervalData` has a many-to-one relationship with it's parent
  interval object through it's ``parent_id`` attribute.

  Args:
    parent_id (int): parent record Interval id
    sample_id (str): unique sample identifier
    group_id (str): group identifier
    coverage (float): average coverage for the exon
    completeness (float): ratio of adequately covered bases
  """

  __tablename__ = 'interval_data'

  id = Column(Integer, primary_key=True, autoincrement=True)
  coverage = Column(Float)
  completeness = Column(Float)

  # These columns map coverage/completeness to sample+group
  sample_id = Column(String(32), ForeignKey('sample.id'))
  sample = relationship(Sample, backref=backref('intervals'))
  group_id = Column(String(32), index=True)

  # Genetic relationship
  parent_id = Column(String(32), ForeignKey('interval.id'))
  parent = relationship(Interval, backref=backref('data'))

  def __init__(self, **kwargs):
    super(IntervalData, self).__init__()

    for key, value in kwargs.items():
      setattr(self, key, value)


# +--------------------------------------------------------------------+
# | Block Data ORM
# +--------------------------------------------------------------------+
class BlockData(Base):

  """Coverage metrics for a single block and a given sample.

  It has a many-to-one relationship with it's parent set object through
  it's ``parent_id`` attribute.

  Args:
    parent_id (int): Set id
    sample_id (str): unique sample identifier
    group_id (str): group identifier
    coverage (float): average coverage for the set
    completeness (float): ratio of adequately covered bases
  """

  __tablename__ = 'block_data'

  id = Column(Integer, primary_key=True, autoincrement=True)
  coverage = Column(Float)
  completeness = Column(Float)

  # These columns map coverage/completeness to an individual+group
  sample_id = Column(String(32), ForeignKey('sample.id'))
  sample = relationship(Sample, backref=backref('blocks'))
  group_id = Column(String(32), index=True)

  # Genetic relationship
  parent_id = Column(String(32), ForeignKey('block.id'))
  parent = relationship(Block, backref=backref('data'))

  def __init__(self, **kwargs):
    super(BlockData, self).__init__()

    for key, value in kwargs.items():
      setattr(self, key, value)


# +--------------------------------------------------------------------+
# | Superset Data ORM classes
# +--------------------------------------------------------------------+
class SuperblockData(Base):

  """Coverage metrics for a single superblock and a given sample.

  It has a many-to-one relationship with it's parent superset object
  through it's ``parent_id`` attribute.

  Args:
    parent_id (int): Superblock id
    sample_id (str): unique sample identifier
    group_id (str): group identifier
    coverage (float): average coverage for the superset
    completeness (float): ratio of adequately covered bases
  """

  __tablename__ = 'superblock_data'

  id = Column(Integer, primary_key=True, autoincrement=True)
  coverage = Column(Float)
  completeness = Column(Float)

  # These columns map coverage/completeness to an individual+group
  sample_id = Column(String(32), ForeignKey('sample.id'))
  sample = relationship(Sample, backref=backref('superblocks'))
  group_id = Column(String(32), index=True)

  # Genetic relationship
  parent_id = Column(String(32), ForeignKey('superblock.id'))
  parent = relationship(Superblock, backref=backref('data'))

  def __init__(self, **kwargs):
    super(SuperblockData, self).__init__()

    for key, value in kwargs.items():
      setattr(self, key, value)



# -*- coding: utf-8 -*-
"""
chanjo.annotator.stages
~~~~~~~~~~~~~~~~~~~~~~~~

Pipeline stages only used by the Chanjo annotator.
"""
from __future__ import absolute_import, unicode_literals
from operator import attrgetter

from toolz import curry, map

from .._compat import text_type
from ..utils import average, completeness

# function for getting the end of an interval (__doc__ is read only)
end_getter = attrgetter('end')


def assign_relative_positions(abs_start, abs_end, overall_start):
  """Return relative positions given the absolute interval positions.

  Bases the relative positions from an overall starting position.

  Args:
    abs_start (int): global start of the interval
    abs_end (int): global end of the interval
    overall_start (int): absolute start of overall group of intervals

  Returns:
    tuple of int: relative start and end positions
  """
  assert abs_start < abs_end, 'Interval must be positive'
  assert overall_start <= abs_start, "Interval must overlap 'overall'"

  rel_start = abs_start - overall_start
  rel_end = abs_end - overall_start

  return rel_start, rel_end


@curry
def calculate_metrics(interval_readDepth, threshold=10):
  """Calculate coverage and completeness for a range of read depths.

  Assumes a continous interval.

  TODO: in the future, Interval will hold coverage, completeness,
  read_depths etc.

  Args:
    interval (BaseInterval): just passed on through
    read_depths (array): :class:`numpy.array` of read depths for
      **each** of the positions
    threshold (int, optional): cutoff to use for the completeness
      filter, defaults to 10

  Returns:
    BaseInterval: the ``interval`` with added coverage and completeness
  """
  # unpack
  interval, read_depths = interval_readDepth

  return interval._replace(
    coverage=average(read_depths),
    completeness=completeness(read_depths, threshold=threshold)
  )

@curry
def comment_sniffer(line, prefix='#'):
  """Determine if a line is a comment.

  Really a quick wrapper around ``str.startswith``.

  Args:
    line (str): line to check if comment
    prefix (str, optional): prefix to identify comments, default: #.

  Returns:
    bool: whether line is a comment or not
  """
  return text_type.startswith(line, prefix)


@curry
def extend_interval(interval, extension=0):
  """Extend interval symetrically.

  Args:
    interval (:class:`Interval`): Interval object
    extension (int, optional): length to extend interval with,
      defaults to 0

  Returns:
    Interval: interval with start and end replaced
  """
  # extend the interval in both directions, symetrically
  return interval._replace(
    start=interval.start - extension,
    end=interval.end + extension,
  )


@curry
def group_intervals(intervals, bp_threshold=1000):
  """Group and return lists of intervals based on the threshold.

  Args:
    intervals (list): list of intervals
    bp_threshold (int, optional): approx. combined length per group

  Yields:
    list of Interval: next group of intervals
  """
  # this is where we store grouped intervals
  group = []
  group_start = None
  group_end = 0
  last_contig = None

  for interval in intervals:
    # update the current combined interval
    # uses 'max' since some intervals overlap others (all we know is
    # that they are sorted on 'start' position)
    group_end = max(group_end, interval.end)

    # if the current combined interval is big enough or a new contig
    # has started
    new_contig = last_contig != interval.contig
    if new_contig or (group_end - group_start) > bp_threshold:
      # push currently grouped intervals
      if group:
        yield group

      # start a new combined interval group
      group = [interval]
      last_contig = interval.contig

      # reset the combined interval bounderies
      group_start, group_end = interval.start, interval.end

    else:
      # append to the current combined interval group
      group.append(interval)

  # push out the last group
  yield group


def merge_intervals(intervals):
  """Return the ends of a list of intervals.

  Args:
    intervals (list): list of intervals

  Returns:
    tuple of int: the beginning and end of the combined interval
  """
  if len(intervals) == 0:
    raise ValueError("'intervals' must contain at least one interval")

  overall_start = intervals[0].start
  overall_end = max(map(end_getter, intervals))

  return overall_start, overall_end


@curry
def prefix(value, sequence):
  """Prepend "value" to a sequence as ``value + sequence``.

  Args:
    value (variable): item to extend base sequence with
    sequence (sequence): extendable base sequence

  Returns:
    sequence: combination as ``value + sequence``
  """
  return value + sequence


@curry
def process_interval_group(bam, interval_group):
  """Fetch read depths for a group of :class:`Interval`s.

  Args:
    bam (BamFile): initialized BamFile instance
    interval_group (list): list of :class:`Interval`

  Yields:
    tuple of Interval, list: Interval and list of read depths
  """
  # contig is expected to be the same for all intervals
  contig = interval_group[0].contig

  # get the start and end of the group interval, 1:1-based
  overall_start, overall_end = merge_intervals(interval_group)

  # get read depths for the whole (full) group interval
  # => input should be 1:1-based
  read_depths = bam(contig, overall_start, overall_end)

  for interval in interval_group:
    # convert to relative positions for the interval
    # => 1:1-based, includes optional extension
    # => ouput can be considered as '0:0-based' (e.g. 1-1=0)
    args = (interval.start, interval.end, overall_start)
    rel_start, rel_end = assign_relative_positions(*args)

    # slice the overall read depth array with the relative coordinates
    # Python expects 0:1-based when slicing => +1 to 'rel_end'
    read_depth_slice = read_depths[rel_start:rel_end + 1]

    yield interval, read_depth_slice



# -*- coding: utf-8 -*-
from chanjo.exporter import fetch_records
from chanjo.store import Store, Interval


def test_fetch_records():
  # set up
  chanjo_db = Store(':memory:')
  chanjo_db.set_up()
  chanjo_db.add(chanjo_db.create(
    'interval',
    interval_id='interval1',
    contig='chr1',
    start=10,
    end=100,
    strand='+'
  ))
  chanjo_db.save()

  columns = (Interval.contig, Interval.start, Interval.end)
  intervals = list(fetch_records(chanjo_db, columns))
  assert len(intervals) == 1
  assert intervals[0] == ('chr1', 10, 100)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import numpy as np
import pytest

from chanjo.annotator import (
  assign_relative_positions,
  calculate_metrics,
  comment_sniffer,
  extend_interval,
  group_intervals,
  merge_intervals,
  prefix,
  process_interval_group
)
from chanjo.depth_reader import BamFile
from chanjo.utils import BaseInterval


def test_assign_relative_positions():
  """Test recalibrating a position relative to a new absolute start."""
  # test at the start of a larger overall interval
  assert assign_relative_positions(0, 10, 0) == (0, 10)

  # test some random interval
  assert assign_relative_positions(55, 170, 42) == (13, 128)

  # test edge case
  assert assign_relative_positions(2131334, 2132671, 2131334) == (0, 1337)

  # test something that shouldn't be valid; reversed interval
  with pytest.raises(AssertionError):
    assign_relative_positions(123, 100, 90)

  # test something that shouldn't be valid; non-inclusive inverval
  with pytest.raises(AssertionError):
    assign_relative_positions(100, 123, 150)


def test_calculate_metrics():
  """Test calculating multiple metrics."""
  # at this point, completeness can only be calculated with a
  # numpy array of read depths
  # read depts, coverage=4, bp_count=8
  read_depths = np.array([0, 5, 3, 3, 4, 5, 5, 7])
  fake_interval = BaseInterval('chr1', 10, 17)
  interval = calculate_metrics((fake_interval, read_depths), threshold=5)

  assert interval[:3] == fake_interval[:3]
  assert interval.coverage == 4
  assert interval.completeness == .5


def test_comment_sniffer():
  """Test sniffing out comment lines."""
  assert comment_sniffer("I'm real data!") == False
  assert comment_sniffer('#comment') == True
  assert comment_sniffer('|another comment', prefix='|') == True


def test_extend_interval():
  """Test extending an interval in both directions."""
  interval = BaseInterval('chr1', 10, 100)

  # test default which is no extension
  assert extend_interval(interval) == interval

  # test simple extension
  extended_interval = extend_interval(interval, extension=5)
  assert extended_interval.start == 5
  assert extended_interval.end == 105


def test_group_intervals():
  """Test aggregating intervals in groups."""
  # intervals on the same contig
  intervals = [
    BaseInterval('1', 10, 100),
    BaseInterval('1', 90, 150),
    BaseInterval('1', 200, 250)
  ]
  grouped_intervals = [intervals[:2], intervals[2:]]
  groups = group_intervals(intervals, bp_threshold=150)
  assert list(groups) == grouped_intervals

  # intervals on multiple contigs
  intervals = [
    BaseInterval('X', 10, 100),
    BaseInterval('Y', 90, 150),
    BaseInterval('Y', 200, 250)
  ]
  grouped_intervals = [intervals[:1], intervals[1:]]
  groups = group_intervals(intervals, bp_threshold=1000)
  assert list(groups) == grouped_intervals


def test_merge_intervals():
  """Test merging multiple intervals."""
  # only makes sense for intervals on the same contig
  intervals = [
    BaseInterval('1', 10, 100),
    BaseInterval('1', 90, 150),
    BaseInterval('1', 200, 250)
  ]
  assert merge_intervals(intervals) == (10, 250)

  # test with tricky overlapping intervals
  intervals = [
    BaseInterval('1', 505, 585),
    BaseInterval('1', 520, 550),
    BaseInterval('1', 545, 580)
  ]
  assert merge_intervals(intervals) == (505, 585)

  # test with 1 interval
  assert merge_intervals([BaseInterval('1', 10, 10)]) == (10, 10)

  # test with 0 intervals
  with pytest.raises(ValueError):
    merge_intervals([])


def test_prefix():
  """Test prefixing values to sequences."""
  assert prefix('chr', '1') == 'chr1'
  assert prefix([10], [11, 12, 13]) == [10, 11, 12, 13]
  assert prefix('', 'X') == 'X'


def test_process_interval_group():
  """Test processing an entire group of intervals."""
  bam = BamFile('tests/fixtures/alignment.bam')
  interval_group = [
    BaseInterval('chr1', 1, 5), BaseInterval('chr1', 10, 13)
  ]

  intervals_coverage = list(process_interval_group(bam, interval_group))

  interval1 = intervals_coverage[0][0]
  read_depths1 = list(intervals_coverage[0][1])
  assert interval1 == interval_group[0]
  assert read_depths1 == [2, 4, 5, 5, 5]

  interval2 = intervals_coverage[1][0]
  read_depths2 = list(intervals_coverage[1][1])
  assert interval2 == interval_group[1]
  assert read_depths2 == [7, 7, 7, 7]



# -*- coding: utf-8 -*-
from chanjo.config import remove_ansi


def test_remove_ansi():
  # Test a general case
  ansi_string = '[\x1b[32m?\x1b[0m] name: '
  assert remove_ansi(ansi_string) == '[?] name: '



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from functools import partial
import re

# regular expression to delete ANSI escape sequences (UNIX color codes)
# ref: http://stackoverflow.com/questions/14693701
ansi_escape = re.compile(r'\x1b[^m]*m')

remove_ansi = partial(ansi_escape.sub, '')
remove_ansi.__doc__ = """Remove ANSI escape characters from a string.

.. code-block:: python

  >>> ansi_string = '[\x1b[32m?\x1b[0m] name: '
  >>> remove_ansi(ansi_string)
  '[?] name: '

Args:
  ansi_string (str): string, possibly containing ANSI escape characters

Returns:
  str: input string *guaranteed* without ANSI escape characters
"""



# -*- coding: utf-8 -*-
"""
chanjo.cli
~~~~~~~~~~~

Command line interface (console entry points). Based on Click_.

.. _Click: http://click.pocoo.org/
"""
from __future__ import absolute_import, unicode_literals
from pkg_resources import iter_entry_points

import click

from . import __version__
from ._compat import text_type
from .config import Config, config_file_name, markup
from .store import Store


@click.group()
@click.option(
  '-c', '--config',
  default=config_file_name,
  type=click.File('w', encoding='utf-8'),
  help='path to config file')
@click.option('--db', type=text_type, help='path/URI of the SQL database')
@click.option(
  '-d', '--dialect',
  type=click.Choice(['sqlite', 'mysql']),
  help='type of SQL database')
@click.version_option(__version__)
@click.pass_context
def cli(context, config, db, dialect):
  """Clinical sequencing coverage analysis tool."""
  # avoid setting global defaults in Click options, do it below when
  # updating the config object
  context.obj = Config(config, markup=markup)

  # global defaults
  db_path = db or context.obj.get('db', 'coverage.sqlite')
  db_dialect = dialect or context.obj.get('dialect', 'sqlite')

  context.db = Store(db_path, dialect=db_dialect)

  # update the context with new defaults from the config file
  context.default_map = context.obj


# add subcommands dynamically to the CLI
for entry_point in iter_entry_points('chanjo.subcommands'):
  cli.add_command(entry_point.load())



# -*- coding: utf-8 -*-

from chanjo.converter import (
  grep,
  extract_intervals,
  merge_related_elements,
  parse_raw_intervals,
  rename_sex_interval
)
from chanjo.utils import BaseInterval


def test_grep():
  string = 'Paul Thomas Anderson is a fantastic director.'
  pattern = 'fantastic'

  assert grep(pattern, string)
  assert grep('Kubrick', string) == False


def test_parse_raw_intervals():
  result = parse_raw_intervals('[11-18, 25-30, 32-35]')
  assert result == [[11, 18], [25, 30], [32, 35]]


def test_extract_intervals():
  record = ['1', 'NC_000001.10', 'SAMD11', '148398', 'CCDS2.2', 'Public',
            '+', '11', '35', '[11-18, 25-30, 32-35]', 'Identical']

  intervals = list(extract_intervals(record))
  interval1 = BaseInterval(
    '1', 11, 18, '1-11-18', 0, '+', ['CCDS2.2'], ['SAMD11']
  )

  assert intervals[0] == interval1


def test_merge_related_elements():
  interval_group = [
    BaseInterval(
      'X', 10, 100,block_ids=['block1'], superblock_ids=['sblock1']),
    BaseInterval(
      'X', 10, 100, block_ids=['block2'], superblock_ids=['sblock1']),
    BaseInterval(
      'X', 10, 100, block_ids=['block3'], superblock_ids=['sblock2'])
  ]

  merged_interval = merge_related_elements(interval_group)

  assert merged_interval[:3] == ('X', 10, 100)
  assert merged_interval.block_ids == ['block1', 'block2', 'block3']
  assert merged_interval.superblock_ids == ['sblock1', 'sblock1', 'sblock2']


def test_rename_sex_interval():
  sex = BaseInterval('X', 11, 111, block_ids=['block1', 'block2'])
  non_sex = BaseInterval('chr22', 4, 55, superblock_ids=['sblock1'])
  alt_sex = BaseInterval('chrY', 99, 921, superblock_ids=['sblock2'])

  # test with sex interval
  renamed_interval = rename_sex_interval(sex)
  assert renamed_interval.block_ids == ['X-block1', 'X-block2']

  # test non-sex interval
  same_interval = rename_sex_interval(non_sex)
  assert same_interval == non_sex

  # test sex interval with alternate contig names
  renamed_interval = rename_sex_interval(alt_sex, sex_contigs=['chrX', 'chrY'])
  assert renamed_interval.superblock_ids == ['chrY-sblock2']



# -*- coding: utf-8 -*-
import json
import warnings

from sqlalchemy import exc as sa_exc
from toolz import pipe
from toolz.curried import map

from .._compat import text_type
from ..utils import split
from .consumers import build_interval_data
from .utils import convert_old_interval_id


def import_bed(chanjo_db, bed_stream):
  """Import the "annotate" output into an existing Chanjo database.

  Args:
    chanjo_db (Store): ``chanjo.store.Store`` instance
    bed_stream (handle): BED-file handle
  """
  # step 1: load header information as JSON (removing the "#" character)
  metadata = json.loads(next(bed_stream)[1:])

  # extract info used later on
  sample_id = metadata['sample_id']
  group_id = metadata['group_id']

  # add a new Sample record using the metadata (don't commit yet)
  chanjo_db.add(chanjo_db.create('sample', **metadata))

  # suppress Decimal/float conversion warnings
  # ref: http://stackoverflow.com/questions/5225780
  with warnings.catch_warnings():
    warnings.simplefilter('ignore', category=sa_exc.SAWarning)
    # step 2: prepare the rest of the input data for consumption
    sequence = pipe(
      bed_stream,
      map(text_type.rstrip),
      map(split(sep='\t'))
    )

    # step 3: consume the interval data from the sequence
    for interval_data in sequence:
      build_interval_data(chanjo_db, sample_id, group_id, interval_data)

    # step 4: commit updates after loading all intervals
    chanjo_db.save()

    # step 5: extend annotations to blocks and superblocks
    return extend_annotations(chanjo_db, sample_id, group_id)


def extend_annotations(chanjo_db, sample_id, group_id):
  """Extend interval annotations to blocks and superblocks.

  Calculates the mean of metrics of related elements.

  Args:
    chanjo_db (Store): instance of class:`chanjo.store.Store`
    sample_id (str): id of sample to extend
    group_id (str): group id of sample to extend

  Returns:
    bool: ``True`` if successful, ``False`` otherwise.
  """
  # extend interval annotations to blocks
  # we commit so the SQL query in the next step works
  chanjo_db.add([chanjo_db.create(
    'block_data',
    parent_id=raw_block[0],
    sample_id=sample_id,
    group_id=group_id,
    coverage=raw_block[1],
    completeness=raw_block[2]
  ) for raw_block in chanjo_db.block_stats(sample_id)]).save()

  # extend annotations to genes
  chanjo_db.add([chanjo_db.create(
    'superblock_data',
    parent_id=raw_superblock[0],
    sample_id=sample_id,
    group_id=group_id,
    coverage=raw_superblock[1],
    completeness=raw_superblock[2]
  ) for raw_superblock in chanjo_db.superblock_stats(sample_id)]).save()


def import_json(chanjo_db, json_stream):
  """Legacy importer for the old JSON output format of the 'annotate' command.
  Deprecated.
  """
  dump = json.load(json_stream)

  annotations = dump['annotations']
  sample_id = dump['sample_id']
  group_id = dump['group_id']

  if dump['splice']:
    extension = 2
  else:
    extension = 0

  # add a Sample entry with metadata
  sample = chanjo_db.create(
    'sample',
    sample_id=sample_id,
    group_id=str(group_id),
    cutoff=dump['cutoff'],
    extension=extension,
    coverage_source=dump['source'],
    element_source=None
  )

  chanjo_db.add(sample)

  # for each of the annotations (intervals)
  chanjo_db.add([chanjo_db.create(
    'interval_data',
    parent_id=convert_old_interval_id(annotation[2]),
    coverage=annotation[0],
    completeness=annotation[1],
    group_id=group_id,
    sample_id=sample_id
  ) for annotation in annotations])

  # commit intervals before proceeding
  # we do this in part to leverage subsequent SQL queries
  chanjo_db.save()

  # extend annotations to blocks and superblocks
  return extend_annotations(chanjo_db, sample_id, group_id)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import click

from .core import export_intervals


@click.command()
@click.option(
  '--header/--no-header', is_flag=True, default=True, help="include headers")
@click.pass_context
def export(context, header):
  """Export an interval BED stream from an existing Chanjo store."""
  # build a new skeleton SQL interval store
  bed_lines = export_intervals(
    chanjo_db=context.parent.db,
    include_header=header
  )

  # reduce/write the BED lines
  for bed_line in bed_lines:
    click.echo(bed_line)



# -*- coding: utf-8 -*-


def fetch_records(db, columns, order_columns=None):
  """Fetch a list of columns from a *single* SQLAlchemy ORM object.

  Also works for a single table. Returns the results as a list of
  tuples.

  Args:
    db (session): ``sqlalchemy.orm.session`` object with a
      ``.query``-method
    columns (list): list of ``sqlalchemy.Column`` from a single ORM
    order_columns (list, optional): model columns to sort by

  Yields:
    tuple: requested columns from a matching record in the SQL query
  """
  # initialize query
  query = db.query(*columns)

  # order the results on request
  if order_columns:
    query = query.order_by(*order_columns)

  # execute the query and yield the result as a list of tuples
  for record in query.all():
    yield record



# -*- coding: utf-8 -*-

import pytest
from chanjo.depth_reader import BamFile


def test_bam():
  bam = BamFile('tests/fixtures/alignment.bam')

  # These are the 65 read depths matching base pairs in 'alignment.bam'
  reference = [2, 4, 5, 5, 5, 5, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
               7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 5, 3, 3, 2,
               2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]

  # Read BAM from position [1,10]
  depths = bam('chr1', 1, 10)

  # Make assertions: we expect the read depths from 1st to 10th position
  # to be included.
  answer = reference[0:10]
  # Compare the list version of the returned ``numpy.array`` to 'answer'
  assert list(depths) == answer

  # Test also an interval that extends beyond the available reads
  depths = bam('chr1', 35, 45)
  assert list(depths) == reference[34:45]

  # Test interval completely outside the available reads
  depths = bam('chr1', 70, 79)
  assert list(depths) == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

  # Test reading a single base
  depths = bam('chr1', 7, 7)
  assert list(depths) == [6.]

  # This is not the current implementation but should perhaps be?
  # How else can you tell this case from returning only zeros.
  # Test submitting a false chromosome Id
  with pytest.raises(ValueError):
    bam('crh1', 10, 20)

  # Test submitting 0-based positions (the 0th doesn't exist)
  with pytest.raises(ValueError):
    bam('chr1', 0, 9)


def test_bam_without_file():
  # Test with non-existent file
  with pytest.raises(OSError):
    BamFile('tests/fixtures/not-here.bam')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import click

from .core import init_db


@click.command()
@click.option(
  '-f', '--force', is_flag=True, help='overwrite database without warning')
@click.argument(
  'in_stream', type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def build(context, in_stream, force):
  """Construct a new skeleton SQL interval store.

  \b
  IN_STREAM: Chanjo-style BED-file with interval definitions
  """
  # build a new skeleton SQL interval store
  init_db(
    chanjo_db=context.parent.db,
    bed_stream=in_stream,
    overwrite=force
  )



# -*- coding: utf-8 -*-
from chanjo.importer import build_interval_data
from chanjo.store import Store


def test_build_interval_data():
  # setup
  chanjo_db = Store(':memory:')
  chanjo_db.set_up()

  sample_id = 'rokinerl'
  group_id = 'group1'
  interval_data = []



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from .core import Store
from .models import (
  Block,
  BlockData,
  Interval,
  IntervalData,
  Interval_Block,
  Sample,
  Superblock,
  SuperblockData
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from .cli import import_data
from .consumers import build_interval_data
from .core import import_bed, import_json
from .utils import convert_old_interval_id



# -*- coding: utf-8 -*-
from chanjo.config import remove_ansi
from chanjo.config.questions import build_prompt


def test_build_prompt():
  # Test simple prompt
  prompt = '[?] name: (Heisenberg) '
  assert remove_ansi(build_prompt('name', '(Heisenberg)')) == prompt

  # Test custom prompt
  prompt = '[?] What is your name? (Heisenberg) '
  built_prompt = build_prompt('What is your name?', '(Heisenberg)')
  assert prompt == remove_ansi(built_prompt)

  # Test highly customized prompt (default not in the end)
  prompt = '[?] I [walk]ed to China '
  built_prompt = build_prompt('I %sed to China', '[walk]')
  assert prompt == remove_ansi(built_prompt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Based on https://github.com/pypa/sampleproject/blob/master/setup.py."""
from __future__ import unicode_literals
# To use a consistent encoding
from codecs import open
import os
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

# Shortcut for building/publishing to Pypi
if sys.argv[-1] == 'publish':
  os.system('python setup.py sdist bdist_wheel upload')
  sys.exit()


# This is a plug-in for setuptools that will invoke py.test
# when you run python setup.py test
class PyTest(TestCommand):

  """Set up the py.test test runner."""

  def finalize_options(self):
    """Set options for the command line."""
    TestCommand.finalize_options(self)
    self.test_args = []
    self.test_suite = True

  def run_tests(self):
    """Execute the test runner command."""
    # Import here, because outside the required eggs aren't loaded yet
    import pytest
    sys.exit(pytest.main(self.test_args))

# Get the long description from the relevant file
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()


setup(
  name='chanjo',

  # Versions should comply with PEP440. For a discussion on
  # single-sourcing the version across setup.py and the project code,
  # see http://packaging.python.org/en/latest/tutorial.html#version
  version='2.0.0-beta4',

  description='Coverage analysis tool for clinical sequencing',
  long_description=long_description,
  # What does your project relate to? Separate with spaces.
  keywords='coverage sequencing clinical exome completeness diagnostics',
  author='Robin Andeer',
  author_email='robin.andeer@gmail.com',
  license='MIT',

  # The project's main homepage
  url='http://www.chanjo.co/',

  packages=find_packages(exclude=('tests*', 'docs', 'examples')),

  # If there are data files included in your packages that need to be
  # installed, specify them here.
  include_package_data=True,
  package_data=dict(
    chanjo=['demo/*']
  ),
  zip_safe=False,

  install_requires=[
    'Click',
    'setuptools',
    'pysam>=0.7.5',
    'toolz',
    'path.py',
    'toml',
    'numpy',
    'sqlalchemy>=0.8.2'
  ],
  # extras_require=dict(
  #   speed=['numpy']
  # ),
  tests_require=[
    'pytest',
  ],
  cmdclass=dict(
    test=PyTest,
  ),

  # To provide executable scripts, use entry points in preference to the
  # "scripts" keyword. Entry points provide cross-platform support and
  # allow pip to create the appropriate form of executable for the
  # target platform.
  entry_points={
    'console_scripts': [
      'chanjo = chanjo.__main__:cli',
      'sex-check = chanjo.sex_checker:sex_check'
    ],
    'chanjo.converters': [
      'ccds = chanjo.converter:ccds_to_bed',
    ],
    'chanjo.subcommands': [
      'annotate = chanjo.annotator:annotate',
      'build = chanjo.builder:build',
      'config = chanjo.config:config',
      'convert = chanjo.converter:convert',
      'demo = chanjo.demo:demo',
      'export = chanjo.exporter:export',
      'import = chanjo.importer:import_data',
      'init = chanjo.config:init',
    ]
  },

  # See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
  classifiers=[
    # How mature is this project? Common values are:
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    'Development Status :: 4 - Beta',

    # Indicate who your project is intended for
    'Intended Audience :: Science/Research',
    'Topic :: Software Development',
    'Topic :: Scientific/Engineering :: Bio-Informatics',

    # Pick your license as you wish (should match "license" above)
    'License :: OSI Approved :: MIT License',

    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',

    'Environment :: Console',
  ],
)



# -*- coding: utf-8 -*-
"""
chanjo.builder.stages
~~~~~~~~~~~~~~~~~~~~~~

Pipeline stages only used by the Chanjo builder.
"""
from toolz import curry

from .._compat import zip, itervalues


def aggregate(interval_payloads):
  """Group elements together based on a given parent id.

  Aggregates elements based on the key (first item in the payload
  tuple). Yields each group at the end of each new contig.

  Args:
    interval_payloads (list of tuple): consist of (key, element)

  Yields:
    list of tuple: where each item consist of (key, element)
  """
  # store created groups
  groups = {}
  # optimize loading (per contig)
  last_contig = None

  for payload in interval_payloads:
    parent_id = payload[0]
    element = payload[1]

    # yield a batch (per contig)
    if last_contig != element.contig:
      for group in itervalues(groups):
        yield group

      # reset
      groups = {}
      last_contig = element.contig

    if parent_id not in groups:
      groups[parent_id] = [payload]

    else:
      groups[parent_id].append(payload)

  # yield also the groups for the last contig
  for group in itervalues(groups):
    yield group


@curry
def build_interval(chanjo_db, interval):
  """Create a new interval and add it to the current session.

  Yields each combination of the new database interval together with
  related block and superblock ids.

  Args:
    chanjo_db (Store): Chanjo store connected to an existing database
    interval (Interval): interval instance

  Yields:
    tuple: block id, new database interval, related superblock id
  """
  # create a brand new interval
  new_interval = chanjo_db.create(
    'interval',
    interval_id=interval.name,
    contig=interval.contig,
    start=interval.start,
    end=interval.end,
    strand=interval.strand
  )

  # add new interval to the current session
  chanjo_db.add(new_interval)

  # yield the new interval for each of the related blocks
  related_ids = zip(interval.block_ids, interval.superblock_ids)
  for block_id, superblock_id in related_ids:
    yield block_id, new_interval, superblock_id


@curry
def build_block(chanjo_db, interval_group):
  """Create a new block and add it to the current session.

  Args:
    chanjo_db (Store): Chanjo store connected to an existing database
    interval_group (list): group of related database intervals

  Returns:
    tuple: superblock id, new database block
  """
  block_id, some_interval, superblock_id = interval_group[0]

  # create a brand new block
  new_block = chanjo_db.create(
    'block',
    block_id=block_id,
    contig=some_interval.contig,
    start=min([payload[1].start for payload in interval_group]),
    end=max([payload[1].end for payload in interval_group]),
    strand=some_interval.strand,
    superblock_id=superblock_id
  )

  # add the new block to the current session
  chanjo_db.add(new_block)

  # relate the intervals to the block
  new_block.intervals = [interval for _, interval, _ in interval_group]

  return superblock_id, new_block


@curry
def build_superblock(chanjo_db, block_group):
  """Create a new superblock and add it to the current session.

  Args:
    chanjo_db (Store): Chanjo store connected to an existing database
    block_group (list): group of related database blocks

  Returns:
    object: new database block
  """
  superblock_id, some_block = block_group[0]

  # create a brand new superblock
  new_superblock = chanjo_db.create(
    'superblock',
    superblock_id=superblock_id,
    contig=some_block.contig,
    start=min([payload[1].start for payload in block_group]),
    end=max([payload[1].end for payload in block_group]),
    strand=some_block.strand
  )

  # add the new superblock to the current session
  chanjo_db.add(new_superblock)

  # relate the blocks to the superblock
  new_superblock.blocks = [block for _, block in block_group]

  return new_superblock



# -*- coding: utf-8 -*-
"""
chanjo.config
~~~~~~~~~~~~~~

Config module with Click integration.
"""
from __future__ import absolute_import
import toml as markup

from .cli import init, config
from .core import Config, init_pipeline
from .questions import ask, build_prompt, questionnaire
from .utils import remove_ansi


config_file_name="%(program)s.%(extension)s" % dict(
  program=__package__,
  extension=markup.__name__  # Works for JSON, TOML, YAML ...
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from pkg_resources import iter_entry_points, load_entry_point

import click
from toolz import pipe
from toolz.curried import map

from ..utils import serialize_interval


@click.command()
@click.option(
  '-a', '--adapter', default='ccds', help='plugin to use for conversion')
@click.option(
  '-l', '--list', 'list_all', is_flag=True, help='show all installed adapters')
@click.argument(
  'in_stream', type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def convert(context, in_stream, adapter, list_all):
  """Convert a reference database file to a Chanjo BED interval file.

  \b
  IN_STREAM: interval reference file (e.g. CCDS database dump)
  """
  if list_all:
    # list the installed converter options
    for entry_point in iter_entry_points('chanjo.converters'):
      # compose and print the message
      segments = dict(
        program=__package__,
        note=click.style('converter', fg='cyan'),
        plugin=entry_point.name
      )
      click.echo("%(program)s %(note)s %(plugin)s" % segments)

  else:
    try:
      # load a single entry point
      converter_pipeline = load_entry_point(
        __package__, 'chanjo.converters', adapter
      )
    except ImportError:
      segments = dict(
        program=__package__,
        note=click.style('error', fg='red'),
        message="No such converter installed: %s" % adapter
      )
      click.echo("%(program)s %(note)s %(message)s" % segments)
      context.abort()

    # execute converter pipeline
    bed_lines = pipe(
      converter_pipeline(in_stream),
      map(serialize_interval(bed=True))     # stringify/bedify
    )

    # reduce/write the BED lines
    for bed_line in bed_lines:
      click.echo(bed_line)



# -*- coding: utf-8 -*-
"""
chanjo.store.core
~~~~~~~~~~~~~~~~~~
"""
from __future__ import absolute_import, division, unicode_literals

from sqlalchemy import create_engine, func
from sqlalchemy.orm import sessionmaker

from .models import (
  Base,
  Superblock,
  Block,
  Interval,
  Interval_Block,
  SuperblockData,
  BlockData,
  IntervalData,
  Sample
)


class Store(object):

  """SQLAlchemy-based database object.

  Bundles functionality required to setup and interact with various
  related genomic interval elements.

  Examples:
    >>> chanjo_db = Store('data/elements.sqlite')
    >>> chanjo_db.set_up()

  .. note::

    For testing pourposes use ``:memory:`` as the ``path`` argument to
    set up in-memory (temporary) database.

  Args:
    uri (str): path/URI to the database to connect to
    dialect (str, optional): connector + type of database:
      'sqlite'/'mysql'
    debug (bool, optional): whether to output logging information

  Attributes:
    uri (str): path/URI to the database to connect to
    engine (class): SQLAlchemy engine, defines what database to use
    session (class): SQLAlchemy ORM session, manages persistance
    query (method): SQLAlchemy ORM query builder method
    classes (dict): bound ORM classes
  """

  def __init__(self, uri, dialect='sqlite', debug=False):
    super(Store, self).__init__()
    self.uri = uri

    # connect to the SQL database
    if dialect == 'sqlite':
      self.engine = create_engine("sqlite:///%s" % uri, echo=debug)

    else:
      # build URI for MySQL containing:
      # <connector>+<sql_type>://<username>:<password>@<server>/<database>
      auth_path = "%(type)s://%(uri)s" % dict(type=dialect, uri=uri)
      self.engine = create_engine(auth_path, pool_recycle=3600, echo=debug)

    # make sure the same engine is propagated to the Base classes
    Base.metadata.bind = self.engine

    # start a session
    self.session = sessionmaker(bind=self.engine)()

    # shortcut to query method
    self.query = self.session.query

    # ORM class shortcuts to enable fetching models dynamically
    self.classes = {
      'superblock': Superblock,
      'block': Block,
      'interval': Interval,
      'interval_block': Interval_Block,
      'superblock_data': SuperblockData,
      'block_data': BlockData,
      'interval_data': IntervalData,
      'sample': Sample
    }

  @property
  def dialect(self):
    """Return database dialect name used for the current connection.

    Dynamic attribute.

    Returns:
      str: name of dialect used for database connection
    """
    return self.engine.dialect.name

  def set_up(self):
    """Initialize a new database with the default tables and columns.

    Returns:
      Store: self
    """
    # create the tables
    Base.metadata.create_all(self.engine)

    return self

  def tare_down(self):
    """Tare down a database (tables and columns).

    Returns:
      Store: self
    """
    # drop/delete the tables
    Base.metadata.drop_all(self.engine)

    return self

  def save(self):
    """Manually persist changes made to various elements. Chainable.

    Returns:
      Store: ``self`` for chainability
    """
    # commit/persist dirty changes to the database
    self.session.commit()

    return self

  def get(self, typ, type_id):
    """Fetch a specific element or ORM class.

    Calls itself recursively when asked to fetch an element.

    Args:
      typ (str): element key or 'class'
      type_id (str): element id or ORM model id

    Returns:
      model: element or ORM class

    Examples:
      >>> gene = db.get('gene', 'GIT1')
    """
    if typ == 'class':
      return self.classes.get(type_id, None)

    # get an ORM class (recursive)
    klass = self.get('class', typ)

    # return the requested element object (or ``None``) if not found
    return self.session.query(klass).get(type_id)

  def find(self, klass_id, query=None, attrs=None):
    """Fetch one or more elements based on the query.

    If the 'query' parameter is a string :meth:`~chanjo.Store.find`
    will fetch one element; just like `get`. If query is a list it will
    match element ids to items in that list and return a list of
    elements. If 'query' is ``None`` all elements of that class will
    be returned.

    Args:
      klass_id (str): type of element to find
      query (str/list, optional): element Id(s)
      attrs (list, optional): list of columns to fetch

    Returns:
      object/list: element(s) from the database
    """
    # get an ORM class
    klass = self.get('class', klass_id)

    if attrs is not None:
      params = [getattr(klass, attr) for attr in attrs]
    else:
      params = (klass,)

    if query is None:
      # return all `klass_id` elements in the database
      return self.session.query(*params).all()

    elif isinstance(query, list):
      # return all `klass_id` elements in the database
      return self.session.query(*params).filter(klass.id.in_(query)).all()

    elif isinstance(query, str):
      # call 'get' to return the single element
      return self.get(klass_id, query)

    else:
      raise ValueError("'query' must be 'None', 'list', or 'str'")

  def add(self, elements):
    """Add one or more new elements and commit the changes. Chainable.

    Args:
      elements (orm/list): new ORM object instance or list of such

    Returns:
      Store: ``self`` for chainability
    """
    if isinstance(elements, Base):
      # Add the record to the session object
      self.session.add(elements)

    elif isinstance(elements, list):
      # Add all records to the session object
      self.session.add_all(elements)

    return self

  def create(self, class_id, *args, **kwargs):
    r"""Create a new instance of an ORM element.

    If attributes are supplied as a tuple they must be in the correct
    order. Supplying a `dict` doesn't require the attributes to be in
    any particular order.

    Args:
      class_id (str): choice between "superblock", "block", "interval"
      \*args (tuple): list the element attributes in the *correct order*
      \**kwargs (dict): element attributes in whatever order you like

    Returns:
      orm: new ORM instance object
    """
    if args:
      # unpack tuple
      return self.get('class', class_id)(*args)

    elif kwargs:
      # unpack dictionary
      return self.get('class', class_id)(**kwargs)

    else:
      raise TypeError('Submit attributes as arguments or keyword arguments')

  def block_stats(self, sample_id):
    """Calculate block level metrics to annotate 'transcripts'.

    Requires all related intervals to already be properly annotated.

    What's happening is we sum read depths and adequately covered bases
    for each interval in a block and then divide those numbers by the
    total interval length of the block.

    .. note::

      Block annotation needs to be carried out before annotating
      superblocks!

    Args:
      sample_id (str): sample Id to match with coverage annotations

    Returns
      list: list of tuples: ``(<block Id>, <coverage>, <completeness>)``
    """
    # length of an interval (in number of bases, hence +1)
    interval_length = Interval.end - Interval.start + 1

    # length of interval * mean coverage
    cum_interval_coverage = interval_length * IntervalData.coverage

    # summed 'cumulative' coverage for all intervals (of a block)
    cum_block_coverage = func.sum(cum_interval_coverage)

    # summed interval lengths of all intervals (of a block)
    total_block_length = func.sum(interval_length)

    # cumulative block coverage divided by total block length
    mean_block_coverage = cum_block_coverage / total_block_length

    # length of interval * mean completeness
    cum_interval_completeness = interval_length * IntervalData.completeness

    # summed cumulative completeness for all intervals (of a block)
    cum_block_completeness = func.sum(cum_interval_completeness)

    # cumulative block completeness divided by total block length
    mean_block_completeness = cum_block_completeness / total_block_length

    # values to fetch
    interval_block_columns = Interval_Block.columns.values()
    interval_id = interval_block_columns[0]
    block_id = interval_block_columns[1]

    return self.query(
      block_id,
      mean_block_coverage,
      mean_block_completeness
    ).join(IntervalData, interval_id==Interval.id).join(IntervalData.parent)\
     .filter(IntervalData.sample_id==sample_id).group_by(block_id)

  def superblock_stats(self, sample_id):
    """Calculate superblock level metrics to annotate genes.

    Requires all related blocks to already be properly annotated.

    What's happening is that we are simply taking the average of the
    metrics on the transcript level and applying that as gene metrics.
    This gives a decent, albeit not perfect, represenation of gene
    level metrics.

    .. note::

      Annotation of transcripts needs to be acomplished before
      annotating genes!

    Args:
      sample_id (str): sample Id to match with coverage annotations

    Returns:
      list: list of tuples: ``(<gene_id>, <coverage>, <completeness>)``
    """
    return self.query(
      Block.superblock_id,
      func.avg(BlockData.coverage),
      func.avg(BlockData.completeness)
    ).join(BlockData, Block.id==BlockData.parent_id)\
     .filter(BlockData.sample_id == sample_id)\
     .group_by(Block.superblock_id)



# -*- coding: utf-8 -*-
"""
chanjo.config.core
~~~~~~~~~~~~~~~~~~~

Works out of the box with any module that exposes the proper 'load' and
'dump' methods will work. Fox example 'toml' and 'yaml'.
"""
from __future__ import absolute_import, unicode_literals
from codecs import open
import json

from click.termui import echo, style
from path import path

from .questions import questionnaire


class Config(dict):
  """Config object to store nested levels of config values.

  Reads values from a file and updates existing defaults. Can also save
  the updates key-value pairs back to a file.

  Customizable to use e.g. TOML, JSON, or YAML - it just requires an
  object with 'load' and 'dump' methods. Uses built in ``json`` module
  by default.

  Args:
    write_handle (file, optional): Handle to possibly existing file for
      writing to.
    defaults (dict, optional): Lowest priority default values
    markup (object, optional): Markup object, default: json
    save_options (dict, optional): Formatting when saving, e.g.
      ``{'indent': 4, 'sort_keys': True}``
  """
  def __init__(self, write_handle=None, defaults=None, markup=json,
               save_options=None):
    super(Config, self).__init__()
    self.user_data = {}
    self.write_handle = write_handle  # Write-enabled
    self.save_options = (save_options or {})

    # Trust the user knows what she's doing
    self.markup = markup

    # Update the defaults with config data
    self.update((defaults or {}))

    if write_handle:
      self.config_path = path(write_handle.name)

      if self.config_path.isfile():
        # Read data from possible config
        with open(self.config_path, encoding='utf-8') as read_handle:
          self.load(read_handle)

  def _resolve_key(self, dot_key, base=None):
    """Resolve a "dot key" (e.g. person.age).

    Private method.

    Args:
      dot_key (str):
    """
    # The key can be provided as a path to nested pairs
    key_parts = dot_key.split('.')
    key_parts.reverse()

    if base is None:
      section = self
    else:
      section = base

    while len(key_parts) > 1:
      key_part = key_parts.pop()
      if key_part not in section:
        section[key_part] = {}

      section = section[key_part]

    return section, key_parts[0]

  def dotget(self, dot_key, default=None, scope=None):
    # if scope is None:
    #   scope = self

    # section, key = self._resolve_key(dot_key, base=scope)

    # return section.get('key', default)
    raise NotImplementedError

  def set(self, dot_key, value, scope=None):
    """Update a config key-value pair."""
    if scope is None:
      scope = self

    section, key = self._resolve_key(dot_key, base=scope)

    # Set key-value pair
    section[key] = value

    return self

  def unset(self, dot_key, scope=None):
    """Delete a key-value pair."""
    scope = (scope or self)
    section, key = self._resolve_key(dot_key, base=scope)

    # Delete key-value pair
    del section[key]

    return self

  def load(self, read_handle):
    try:
      self.user_data.update(self.markup.load(read_handle))

      self.update(**self.user_data)

    except AttributeError:
      raise NotImplementedError(
        "Markup (%s) must expose a 'load'-method" % str(self.markup)
      )

    except ValueError as ex:
      raise ValueError("Bad syntax: %s" % ex)

    return self

  def save(self, **options):
    """Save the current key-value pairs using the write handle.

    Args:
      options (kwargs, optional): Options to pass to dump (like 'indent')

    Returns:
      Config: returns itself for chainability
    """
    options = (options or self.save_options)
    self.markup.dump(self.user_data, self.write_handle, **options)

    return self


def init_pipeline(program, config, questions):
  """Initializes a config object by interactively asking questions to a
  user. Non-pure."""
  if config.user_data:
    # Some existing user settings were found, warn about overwriting them
    message = "%(program)s %(note)s\tThe existing %(file)s will be updated"
    segments = dict(
      program=program,
      note=style('existing', fg='yellow'),
      file=style(config.config_path.basename(), fg='white')
    )

    echo(message % segments)

  # Launch questionnaire
  user_defaults = questionnaire(questions)

  # Set the selected user defaults
  for dot_key, value in user_defaults.items():
    config.set(dot_key, value, scope=config.user_data)

  # Write to the config file
  config.save()



# -*- coding: utf-8 -*-
from chanjo.importer import convert_old_interval_id


def test_convert_old_interval_id():
  # test simple case, chromosome 1, position 10 -> 15
  assert convert_old_interval_id('1-10-15') == '1-11-16'

  # test with 'chr' prefix on chromosome id
  assert convert_old_interval_id('chrX-588234-588847') == 'chrX-588235-588848'

  # test edge case with 0
  assert convert_old_interval_id('Y-0-1201') == 'Y-1-1202'



# -*- coding: utf-8 -*-
from chanjo.exporter import export_intervals
from chanjo.store import Store


def test_exporter_pipeline():
  # set up
  chanjo_db = Store(':memory:')
  chanjo_db.set_up()
  chanjo_db.add(chanjo_db.create(
    'interval',
    interval_id='intervalY',
    contig='Y',
    start=123,
    end=501,
    strand='-'
  ))
  chanjo_db.save()

  bed_lines = list(export_intervals(chanjo_db))

  assert bed_lines[0] == '#chrom\tchromStart\tchromEnd\tname\tscore\tstrand'
  assert bed_lines[1:] == ['Y\t122\t501\tintervalY\t0\t-']



# -*- coding: utf-8 -*-
"""
chanjo.read_depth.core
~~~~~~~~~~~~~~~~~~~~~~~
Closure for interfacing directly with BAM alignment files. Uses
:class:`Samfile` from pysam. It's used by Chanjo to read, read depths
across various genomics intervals.

Depends on the Pysam_ package and requires Samtools_ to be installed in
your ``$PATH`` to work. N.B. by installing *Pysam* through *pip*,
*Samtools* will by installed alongside.

.. _Pysam: http://www.cgat.org/~andreas/documentation/pysam/contents.html
.. _Samtools: http://samtools.sourceforge.net/
"""
from __future__ import absolute_import, unicode_literals
import errno
import os

from pysam import Samfile
from toolz import partial

try:
  import numpy as np
  # numpy dtype to use for read depths
  prealloc_func = partial(np.zeros, dtype=np.int_)

except ImportError:
  def prealloc_func(size, default=0):
    """Preallocate a list of "default"."""
    return [default] * size


def BamFile(bam_path):
  """Return enclosed function to read, read depths from the "bam_path".

  .. code-block:: python

    >>> from chanjo.read_depth import BamFile
    >>> read_depths = BamFile('./alignment.bam')

  Args:
    bam_path (path): path to alignment BAM-file

  Returns:
    function: function to read from the BAM-file
  """
  # raise an error if the file doesn't exist
  if not os.path.exists(bam_path):
    raise OSError(errno.ENOENT, bam_path)

  bam = Samfile(bam_path)
  try:
    bam.pileup()
  except ValueError:
    # catch error when BAM-file isn't indexed (+ ".bai" file)
    raise OSError(
      errno.ENOENT,
      "BAM-file (%s) must be indexed." % os.path.basename(bam_path)
    )

  def reader(contig, start, end):
    """Generate a list of read depths for each position (start, end).

    The `numpy` array is used to optimize performance when building and
    slicing the list.

    This function depends on `Pysam` >=0.7.5 since the ``truncate``
    option wasn't available in previous versions.

    .. code-block:: python

      >>> bamfile.read('17', 1, 5)
      array([3., 4., 4., 5., 4.])

    .. note::

      Positions are expected to be 1:1-based. In other words; if
      start=1, end=9 you should expect read depths for base pair
      positions 1-9 to be returned.

    Args:
      contig (str): contig/chromosome id (str) of interest
      start (int): first position of the interval (1-based)
      end (int): last position of the interval (1-based)

    Returns:
      list or numpy.array: array of read depths for *each* position in
        the interval
    """
    # convert start to 0-based since this is what pysam expects!
    pysam_start = start - 1

    # pysam expects contig as bytes in Python 2
    pysam_contig = str(contig)

    # check that we don't have a negative start position
    if pysam_start < 0:
      raise ValueError("Start position must be > 0, not %d" % start)

    # preallocate an array of 0 read depth for each position
    # pysam excludes positions with 0 read depth
    read_depths = prealloc_func(end - pysam_start)

    try:
      # overwrite read-covered positions (>0 read depth)
      # ``truncate`` ensures it starts and ends on the gives positions
      # note: ``col.pos`` is 0-based, as is ``pysam_start``
      for col in bam.pileup(pysam_contig, pysam_start, end, truncate=True):
        read_depths[col.pos - pysam_start] = col.n

    except ValueError as ve:
      # catch errors where the contig doesn't exist in the BAM-file
      raise ValueError(
        "Must use contig that exist in the Bam-file. Error: %s" % ve)

    return read_depths

  return reader



# -*- coding: utf-8 -*-
"""
chanjo._compat
~~~~~~~~~~~~~~~

Python 2.7.x, 3.2+ compatability module.
"""
from __future__ import absolute_import, unicode_literals
import operator
import sys

is_py2 = sys.version_info[0] == 2


if not is_py2:
  # Python 3
  # strings and ints
  text_type = str
  string_types = (str,)
  integer_types = (int,)

  # lazy iterators
  zip = zip
  range = range
  iteritems = operator.methodcaller('items')
  iterkeys = operator.methodcaller('keys')
  itervalues = operator.methodcaller('values')

else:
  # Python 2
  # strings and ints
  text_type = unicode
  string_types = (str, unicode)
  integer_types = (int, long)

  # lazy iterators
  range = xrange
  from itertools import izip as zip
  iteritems = operator.methodcaller('iteritems')
  iterkeys = operator.methodcaller('iterkeys')
  itervalues = operator.methodcaller('itervalues')
